# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: dosna_transport
   :platform: Unix
   :synopsis: Transports data using DosNa(which provides several storing \
   backends) at the beginning and end of the process list respectively.

.. moduleauthor:: Emilio Perez Juarez <scientificsoftware@diamond.ac.uk>

"""

import logging

from savu.core.transport_setup import MPI_setup
from savu.core.transports.base_transport import BaseTransport
from savu.core.transports.hdf5_transport import Hdf5Transport
from savu.data.chunking import Chunking
from savu.plugins.savers.utils.hdf5_utils import Hdf5Utils

import dosna as dn

log = logging.getLogger(__name__)

DEFAULT_CONNECTION = "savu-data"
DEFAULT_BACKEND = "ceph"
DEFAULT_ENGINE = "mpi"

class DosnaTransport(BaseTransport):
    """Transport implementation to use DosNa for managing storage and
    chunking"""

    def __init__(self):
        super(DosnaTransport, self).__init__()
        self.dosna_connection = None
        self.global_data = True
        self.h5trans = Hdf5Transport()
        self.data_flow = None
        self.count = 0
        self.hdf5 = None
        self.hdf5_flag = True
        self.files = []
        self.final_dict = None
        self.dataset_cache = []
        self.n_plugins = 0

    def _transport_initialise(self, options):
        MPI_setup(options)

        backend = options.get("dosna_backend") or DEFAULT_BACKEND
        engine = options.get("dosna_engine") or DEFAULT_ENGINE
        dosna_connection_name = options.get("dosna_connection") \
            or DEFAULT_CONNECTION
        dosna_connection_options = options.get("dosna_connection_options")

        dosna_options = {}

        dosna_options.update(dict(item.split('=')
                             for item in dosna_connection_options))
        log.debug("DosNa is using backend %s engine %s and options %s",
                  backend, engine, dosna_options)
        dn.use(engine, backend)
        self.dosna_connection = dn.Connection(dosna_connection_name,
                                              **dosna_options)
        self.dosna_connection.connect()
        # initially reading from a hdf5 file so Hdf5TransportData will be used
        # for all datasets created in a loader
        options['transport'] = 'hdf5'

    def _transport_update_plugin_list(self):
        plugin_list = self.exp.meta_data.plugin_list
        saver_idx = plugin_list._get_savers_index()
        remove = []

        # check the saver plugin and turn off if it is hdf5
        for idx in saver_idx:
            if plugin_list.plugin_list[idx]['name'] == 'Hdf5Saver':
                remove.append(idx)
        for idx in sorted(remove, reverse=True):
            plugin_list._remove(idx)

    def _transport_pre_plugin_list_run(self):
        # loaders have completed now revert back to DosnaTransport, so any
        # output datasets created by a plugin will use this.
        self.hdf5 = Hdf5Utils(self.exp)
        exp_coll = self.exp._get_collection()
        self.data_flow = self.exp.meta_data.plugin_list._get_dataset_flow()
        #self.exp.meta_data.set('transport', 'dosna')
        plist = self.exp.meta_data.plugin_list
        self.n_plugins = plist._get_n_processing_plugins()
        self.final_dict = plist.plugin_list[-1]
        for plugin_index in range(self.n_plugins):
            self.exp._set_experiment_for_current_plugin(plugin_index)
            self.files.append(
                self._get_filenames(exp_coll['plugin_dict'][plugin_index]))
            self._set_file_details(self.files[plugin_index])
            self._setup_dosna_objects()  # creates the dosna objects

        if self.n_plugins is not 1:
            self.exp.meta_data.set('transport', 'dosna')

    def _transport_post_plugin_list_run(self):
        if not self.dosna_connection:
            return
        for dataset in self.dataset_cache:
            self.dosna_connection.del_dataset(dataset.name)
        self.dataset_cache = []
        self.dosna_connection.disconnect()
        self.dosna_connection = None

    def _transport_terminate_dataset(self, data):
        if self.exp.meta_data.get('transport') == "hdf5":
            self.hdf5._close_file(data)

    @staticmethod
    def _extract_digits(data):
        result = []
        for char in data:
            if ord(char) in range(ord('0'), ord('9') + 1):
                result.append(char)
        return "".join(result)

    def _create_dosna_dataset(self, object_id, data, key, current_and_next):
        group_name = self.exp.meta_data.get(["group_name", key])
        data.data_info.set('group_name', group_name)
        try:
            group_name = group_name + '_' + data.name
        except AttributeError:
            pass

        shape = data.get_shape()
        dataset_name = "{}_{}".format(group_name,
                                      self._extract_digits(object_id))

        if current_and_next is 0:
            data.data = self.dosna_connection.create_dataset(dataset_name,
                                                             shape,
                                                             data.dtype)
        else:
            chunking = Chunking(self.exp, current_and_next)
            chunks = chunking._calculate_chunking(shape, data.dtype)
            data.data = self.dosna_connection.create_dataset(dataset_name,
                                                             shape,
                                                             data.dtype,
                                                             chunk_size=chunks)
        self.dataset_cache.append(data.data)

    def _setup_dosna_objects(self):
        out_data_dict = self.exp.index["out_data"]

        current_and_next = [0]*len(out_data_dict)
        if 'current_and_next' in self.exp.meta_data.get_dictionary():
            current_and_next = self.exp.meta_data.get('current_and_next')

        for key in list(out_data_dict.keys()):
            out_data = out_data_dict[key]
            filename = self.exp.meta_data.get(["filename", key])
            self._create_dosna_dataset(filename, out_data, key,
                                       current_and_next[key])

    def _transport_pre_plugin(self):
        if self.count == self.n_plugins - 1:
            self.__set_hdf5_transport()

    def _transport_post_plugin(self):
        # revert back to basic if a temporary transport mechanism was used
        if self.hdf5_flag:
            self.__unset_hdf5_transport()

        if self.count == self.n_plugins - 2:
            self.exp.meta_data.set('transport', 'dosna')

        if self.count == self.n_plugins - 1:  # final plugin
            self.h5trans.exp = self.exp
            self.h5trans.hdf5 = Hdf5Utils(self.exp)
            self.h5trans._transport_post_plugin()

        self.count += 1

    def __set_hdf5_transport(self):
        self.hdf5_flag = True
        self.exp.meta_data.set('transport', 'hdf5')
        files = self._get_filenames(self.final_dict)
        self._set_file_details(files)
        self._setup_h5_files()

    def __unset_hdf5_transport(self):
        self.exp.meta_data.set('transport', 'basic')
        self.hdf5_flag = False



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: histogram
   :platform: Unix
   :synopsis: histograms a binary input

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.plugins.analysis.base_analysis import BaseAnalysis
import numpy as np


@register_plugin
class Histogram(BaseAnalysis, CpuPlugin):
    """
    This plugin uses peakutils to find peaks in spectra. This is then metadata.
    """

    def __init__(self):
        super(Histogram, self).__init__("Histogram")

    def process_frames(self, data):
        return np.array([data[0].sum()])
    #TODO: probably should add the option for some filtering on the output/ rebinning etc

    def setup(self):
        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()
        # set information relating to the plugin data
        in_pData, out_pData = self.get_plugin_datasets()

        # set pattern_name and nframes to process for all datasets
        in_pData[0].plugin_data_setup("CHANNEL", self.get_max_frames())
        sh = in_dataset[0].get_shape()
        axis_labels = ['idx.unit', 'peaks.pixels']
        out_dataset[0].create_dataset(axis_labels=axis_labels,
                                      shape=(sh[-1], 1))
        out_dataset[0].add_pattern("METADATA", slice_dims=(0,), core_dims=(1,))
        out_pData[0].plugin_data_setup("METADATA", self.get_max_frames())

    def get_max_frames(self):
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: vo_centering
   :platform: Unix
   :synopsis: A plugin to find the center of rotation per frame
.. moduleauthor:: Mark Basham, Nghia Vo, Nicola Wadeson \
                    <scientificsoftware@diamond.ac.uk>
"""
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
from savu.data.plugin_list import CitationInformation
import savu.core.utils as cu

import logging
import numpy as np
import scipy.ndimage as ndi
import pyfftw.interfaces.scipy_fftpack as fft



import sys

@register_plugin
class VoCentering(BaseFilter, CpuPlugin):
    """
    A plugin to calculate the centre of rotation using the Vo Method
    :u*param preview: A slice list of required frames (sinograms) to use in \
    the calulation of the centre of rotation (this will not reduce the data \
    size for subsequent plugins). Default: [].
    :u*param start_pixel: The estimated centre of rotation. If value is None,\
        use the horizontal centre of the image. Default: None.
    :u*param search_area: Search area around the estimated centre of rotation\
        . Default: (-50, 50).
    :u*param ratio: The ratio between the size of object and FOV of \
        the camera. Default: 0.5.
    :param search_radius: Use for fine searching. Default: 6.
    :param step: Step of fine searching. Default: 0.5.
    :param datasets_to_populate: A list of datasets which require this \
        information. Default: [].
    :param out_datasets: The default names\
        . Default: ['cor_preview','cor_broadcast'].
    :param broadcast_method: Method of broadcasting centre values calculated\
        from preview slices to full dataset. Available option: 'median', \
        'mean', 'nearest', 'linear_fit'. Default: 'median'.
    :param row_drop: Drop lines around vertical center of the \
        mask. Default: 20.
    :param average_radius: Averaging sinograms around a required sinogram to\
        improve signal-to-noise ratio. Default: 5.
    """

    def __init__(self):
        super(VoCentering, self).__init__("VoCentering")

    def _create_mask(self, nrow, ncol, radius, drop):
        du = 1.0 / ncol
        dv = (nrow - 1.0) / (nrow * 2.0 * np.pi)
        cen_row = np.int16(np.ceil(nrow / 2.0) - 1)
        cen_col = np.int16(np.ceil(ncol / 2.0) - 1)
        drop = min(drop, np.int16(np.ceil(0.05 * nrow)))
        mask = np.zeros((nrow, ncol), dtype='float32')
        for i in range(nrow):
            pos = np.int16(np.round(((i - cen_row) * dv / radius) / du))
            (pos1, pos2) = np.clip(np.sort(
                (-pos + cen_col, pos + cen_col)), 0, ncol - 1)
            mask[i, pos1:pos2 + 1] = 1.0
        mask[cen_row - drop:cen_row + drop + 1,:] = 0.0
        mask[:, cen_col-1:cen_col+2] = 0.0
        return mask

    def _coarse_search(self, sino, start_cor, stop_cor, ratio, drop):
        """
        Coarse search for finding the rotation center.
        """
        (nrow, ncol) = sino.shape
        start_cor, stop_cor = np.sort((start_cor,stop_cor))
        start_cor = np.int16(np.clip(start_cor, 0, ncol-1))
        stop_cor = np.int16(np.clip(stop_cor, 0, ncol-1))
        cen_fliplr = (ncol - 1.0) / 2.0
        # Flip left-right the [0:Pi ] sinogram to make a full [0;2Pi] sinogram
        flip_sino = np.fliplr(sino)
        # Below image is used for compensating the shift of the [Pi;2Pi] sinogram
        # It helps to avoid local minima.
        comp_sino = np.flipud(sino)
        list_cor = np.arange(start_cor, stop_cor + 1.0)
        list_metric = np.zeros(len(list_cor), dtype=np.float32)
        mask = self._create_mask(2 * nrow, ncol, 0.5 * ratio * ncol, drop)
        sino_sino = np.vstack((sino, flip_sino))
        for i, cor in enumerate(list_cor):
            shift = np.int16(2.0*(cor - cen_fliplr))
            _sino = sino_sino[nrow:]
            _sino[...] = np.roll(flip_sino, shift, axis=1)
            if shift >= 0:
                _sino[:, :shift] = comp_sino[:, :shift]
            else:
                _sino[:, shift:] = comp_sino[:, shift:]
            list_metric[i] = np.mean(
                np.abs(np.fft.fftshift(fft.fft2(sino_sino)))*mask)
        minpos = np.argmin(list_metric)
        if minpos==0:
            self.error_msg_1 = "!!! WARNING !!! Global minimum is out of "\
            "the searching range. Please extend smin"
            logging.warning(self.error_msg_1)
            cu.user_message(self.error_msg_1)
        if minpos==len(list_metric)-1:
            self.error_msg_2 = "!!! WARNING !!! Global minimum is out of "\
             "the searching range. Please extend smax"
            logging.warning(self.error_msg_2)
            cu.user_message(self.error_msg_2)
        rot_centre = list_cor[minpos]
        return rot_centre

    def _fine_search(self, sino, start_cor, search_radius,
                      search_step, ratio, drop):
        """
        Fine search for finding the rotation center.
        """
        # Denoising
        (nrow, ncol) = sino.shape
        flip_sino = np.fliplr(sino)
        search_radius = np.clip(np.abs(search_radius), 1, ncol//10 - 1)
        search_step = np.clip(np.abs(search_step), 0.1, 1.1)
        start_cor = np.clip(start_cor, search_radius, ncol - search_radius - 1)
        cen_fliplr = (ncol - 1.0) / 2.0
        list_cor = start_cor + np.arange(
                -search_radius, search_radius + search_step, search_step)
        comp_sino = np.flipud(sino) # Used to avoid local minima
        list_metric = np.zeros(len(list_cor), dtype = np.float32)
        mask = self._create_mask(2 * nrow, ncol, 0.5 * ratio * ncol, drop)
        for i, cor in enumerate(list_cor):
            shift = 2.0*(cor - cen_fliplr)
            sino_shift = ndi.interpolation.shift(
                flip_sino, (0, shift), order = 3, prefilter = True)
            if shift>=0:
                shift_int = np.int16(np.ceil(shift))
                sino_shift[:,:shift_int] = comp_sino[:,:shift_int]
            else:
                shift_int = np.int16(np.floor(shift))
                sino_shift[:,shift_int:] = comp_sino[:,shift_int:]
            mat1 = np.vstack((sino, sino_shift))
            list_metric[i] = np.mean(
                np.abs(np.fft.fftshift(fft.fft2(mat1)))*mask)
        min_pos = np.argmin(list_metric)
        cor = list_cor[min_pos]
        return cor

    def _downsample(self, image, dsp_fact0, dsp_fact1):
        """
        Downsample an image by averaging.
        ---------
        Parameters: - image: 2D array.
                    - dsp_fact0: downsampling factor along axis 0.
                    - dsp_fact1: downsampling factor along axis 1.
        ---------
        Return:     - Downsampled image.
        """
        (height, width) = image.shape
        dsp_fact0 = np.clip(np.int16(dsp_fact0), 1, height // 2)
        dsp_fact1 = np.clip(np.int16(dsp_fact1), 1, width // 2)
        height_dsp = height//dsp_fact0
        width_dsp = width//dsp_fact1
        if dsp_fact0 == 1 and dsp_fact1 == 1:
            image_dsp = image
        else:
            image_dsp = image[0:dsp_fact0*height_dsp,0:dsp_fact1*width_dsp]
            image_dsp = image_dsp.reshape(
                height_dsp,dsp_fact0,width_dsp,dsp_fact1).mean(-1).mean(1)
        return image_dsp

    def set_filter_padding(self, in_data, out_data):
        padding = np.int16(self.parameters['average_radius'])
        if padding>0:
            in_data[0].padding = {'pad_multi_frames': padding}

    def pre_process(self):
        self.drop = np.int16(self.parameters['row_drop'])
        self.smin, self.smax = np.int16(self.parameters['search_area'])
        self.search_radius = np.float32(self.parameters['search_radius'])
        self.search_step = np.float32(self.parameters['step'])
        self.ratio = np.float32(self.parameters['ratio'])
        self.est_cor = self.parameters['start_pixel']
        self.broadcast_method = str(self.parameters['broadcast_method'])
        self.error_msg_1 = ""
        self.error_msg_2 = ""
        self.error_msg_3 = ""
        if not((self.broadcast_method == 'mean')
                or (self.broadcast_method == 'median')
                 or (self.broadcast_method == 'linear_fit')
                  or (self.broadcast_method == 'nearest')):
            self.error_msg_3 = "!!! WARNING !!! Selected broadcasting "\
             "method is out of the list. Use the default option: 'median'"
            logging.warning(self.error_msg_3)
            cu.user_message(self.error_msg_3)
            self.broadcast_method = 'median'
        in_pData = self.get_plugin_in_datasets()[0]
        data = self.get_in_datasets()[0]
        starts,stops,steps = data.get_preview().get_starts_stops_steps()[0:3]
        start_ind = starts[1]
        stop_ind = stops[1]
        step_ind = steps[1]
        name = data.get_name()
        pre_start = self.exp.meta_data.get(name + '_preview_starts')[1]
        pre_stop = self.exp.meta_data.get(name + '_preview_stops')[1]
        pre_step = self.exp.meta_data.get(name + '_preview_steps')[1]
        self.origin_prev = np.arange(pre_start,pre_stop, pre_step)
        self.plugin_prev = self.origin_prev[start_ind:stop_ind:step_ind]

    def process_frames(self, data):
        if len(data[0].shape)>2:
            sino = np.mean(data[0],axis=1)
        else:
            sino = data[0]
        (nrow, ncol) = sino.shape
        dsp_row = 1
        dsp_col = 1
        if ncol>2000:
            dsp_col = 4
        if nrow>2000:
            dsp_row = 2
        # Denoising
        # There's a critical reason to use different window sizes
        # between coarse and fine search.
        sino_csearch = ndi.gaussian_filter(sino, (3,1), mode='reflect')
        sino_fsearch = ndi.gaussian_filter(sino, (2,2), mode='reflect')
        sino_dsp = self._downsample(sino_csearch, dsp_row, dsp_col)
        fine_srange = max(self.search_radius, dsp_col)
        off_set = 0.5*dsp_col if dsp_col > 1 else 0.0
        if self.est_cor is None:
            self.est_cor = (ncol-1.0) / 2.0
        else:
            self.est_cor = np.float32(self.est_cor)
        start_cor = np.int16(
            np.floor(1.0 * (self.est_cor + self.smin) / dsp_col))
        stop_cor = np.int16(
            np.ceil(1.0 * (self.est_cor + self.smax) / dsp_col))
        raw_cor = self._coarse_search(sino_dsp, start_cor, stop_cor,
                                       self.ratio, self.drop)
        cor = self._fine_search(
            sino_fsearch, raw_cor*dsp_col + off_set, fine_srange,
             self.search_step, self.ratio, self.drop)
        return [np.array([cor]), np.array([cor])]

    def post_process(self):
        in_datasets, out_datasets = self.get_datasets()
        cor_prev = out_datasets[0].data[...]
        cor_broad = out_datasets[1].data[...]
        cor_broad[:] = np.median(np.squeeze(cor_prev))
        self.cor_for_executive_summary = np.median(cor_broad[:])
        if self.broadcast_method == 'mean':
            cor_broad[:] = np.mean(np.squeeze(cor_prev))
            self.cor_for_executive_summary = np.mean(cor_broad[:])
        if (self.broadcast_method == 'linear_fit') and (len(cor_prev)>1):
            afact, bfact = np.polyfit(self.plugin_prev, cor_prev[:,0], 1)
            list_cor = self.origin_prev*afact + bfact
            cor_broad[:,0] = list_cor
            self.cor_for_executive_summary = cor_broad[:]
        if (self.broadcast_method == 'nearest') and (len(cor_prev)>1):
            for i, pos in enumerate(self.origin_prev):
                minpos = np.argmin(np.abs(pos-self.plugin_prev))
                cor_broad[i,0] = cor_prev[minpos,0]
            self.cor_for_executive_summary = cor_broad[:]
        out_datasets[1].data[:] = cor_broad[:]
        self.populate_meta_data('cor_preview', np.squeeze(cor_prev))
        self.populate_meta_data('centre_of_rotation',
                                out_datasets[1].data[:].squeeze(axis=1))

    def populate_meta_data(self, key, value):
        datasets = self.parameters['datasets_to_populate']
        in_meta_data = self.get_in_meta_data()[0]
        in_meta_data.set(key, value)
        for name in datasets:
            self.exp.index['in_data'][name].meta_data.set(key, value)

    def setup(self):
        self.exp.log(self.name + " Start calculating center of rotation")
        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', self.get_max_frames())
        slice_dirs = list(in_dataset[0].get_slice_dimensions())
        self.orig_full_shape = in_dataset[0].get_shape()

        # if preview parameters exist then use these
        # else get the size of the data
        # get n processes and take 4 different sets of 5 from the data if this is feasible based on the data size.
        # calculate the slice list here and determine if it is feasible, else apply to max(n_processes, data_size)

        # reduce the data as per data_subset parameter
        self.set_preview(in_dataset[0], self.parameters['preview'])
        total_frames = \
            self._calc_total_frames(in_dataset[0].get_preview(), slice_dirs)

        # copy all required information from in_dataset[0]
        fullData = in_dataset[0]
        new_shape = (np.prod(np.array(fullData.get_shape())[slice_dirs]), 1)
        self.orig_shape = \
            (np.prod(np.array(self.orig_full_shape)[slice_dirs]), 1)
        out_dataset[0].create_dataset(shape=new_shape,
                                      axis_labels=['x.pixels', 'y.pixels'],
                                      remove=True,
                                      transport='hdf5')
        out_dataset[0].add_pattern("METADATA", core_dims=(1,), slice_dims=(0,))

        out_dataset[1].create_dataset(shape=self.orig_shape,
                                      axis_labels=['x.pixels', 'y.pixels'],
                                      remove=True,
                                      transport='hdf5')
        out_dataset[1].add_pattern("METADATA", core_dims=(1,), slice_dims=(0,))
        out_pData[0].plugin_data_setup('METADATA', self.get_max_frames())
        out_pData[1].plugin_data_setup('METADATA', self.get_max_frames())
        out_pData[1].meta_data.set('fix_total_frames', total_frames)
        self.exp.log(self.name + " End")

    def _calc_total_frames(self, preview, slice_dims):
        starts, stops, steps, _ = preview.get_starts_stops_steps()
        lengths = [len(np.arange(starts[i], stops[i], steps[i]))
                   for i in range(len(starts))]
        return np.prod([lengths[i] for i in slice_dims])

    def nOutput_datasets(self):
        return 2

    def get_max_frames(self):
        return 'single'

    def fix_transport(self):
        return 'hdf5'

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The center of rotation for this reconstruction was calculated " +
             "automatically using the method described in this work")
        cite_info.bibtex = \
            ("@article{vo2014reliable,\n" +
             "title={Reliable method for calculating the center of rotation " +
             "in parallel-beam tomography},\n" +
             "author={Vo, Nghia T and Drakopoulos, Michael and Atwood, " +
             "Robert C and Reinhard, Christina},\n" +
             "journal={Optics Express},\n" +
             "volume={22},\n" +
             "number={16},\n" +
             "pages={19078--19086},\n" +
             "year={2014},\n" +
             "publisher={Optical Society of America}\n" +
             "}")
        cite_info.endnote = \
            ("%0 Journal Article\n" +
             "%T Reliable method for calculating the center of rotation in " +
             "parallel-beam tomography\n" +
             "%A Vo, Nghia T\n" +
             "%A Drakopoulos, Michael\n" +
             "%A Atwood, Robert C\n" +
             "%A Reinhard, Christina\n" +
             "%J Optics Express\n" +
             "%V 22\n" +
             "%N 16\n" +
             "%P 19078-19086\n" +
             "%@ 1094-4087\n" +
             "%D 2014\n" +
             "%I Optical Society of America")
        cite_info.doi = "https://doi.org/10.1364/OE.22.019078"
        return cite_info

    def executive_summary(self):
        if ((self.error_msg_1 == "")
             and (self.error_msg_2 == "")):
            msg = "Centre of rotation is : %s" % (str(self.cor_for_executive_summary))
        else:
            msg = "\n" + self.error_msg_1 + "\n" + self.error_msg_2
            msg2 = "(Not well) estimated centre of rotation is : %s" % (str(
                self.cor_for_executive_summary))
            cu.user_message(msg2)
        return [msg]


# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: raven_filter
   :platform: Unix
   :synopsis: A plugin to remove ring artefacts

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""
import logging
import numpy as np
import pyfftw
import pyfftw.interfaces.numpy_fft as fft

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.data.plugin_list import CitationInformation
from savu.plugins.utils import register_plugin


@register_plugin
class RavenFilter(BaseFilter, CpuPlugin):
    """
    Ring artefact removal method

    :u*param uvalue: To define the shape of filter, e.g. bad=10, moderate=20,\
        minor=50. Default: 20.
    :param vvalue: How many rows to be applied the filter. Default: 2.
    :param nvalue: To define the shape of filter. Default: 4.
    :param padFT: Padding for Fourier transform. Default: 20.
    """

    def __init__(self):
        logging.debug("Starting Raven Filter")
        super(RavenFilter, self).__init__("RavenFilter")
        self.count = 0

    def set_filter_padding(self, in_data, out_data):
        self.pad = self.parameters['padFT']
        # don't currently have functionality to pad top/bottom but not
        # right/left so padding everywhere for now
        in_data[0].padding = {'pad_frame_edges': self.pad}
        out_data[0].padding = {'pad_frame_edges': self.pad}

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()[0]
        sino_shape = list(in_pData.get_shape())

        width1 = sino_shape[1] + 2*self.pad
        height1 = sino_shape[0] + 2*self.pad

        v0 = np.abs(self.parameters['vvalue'])
        u0 = np.abs(self.parameters['uvalue'])
        n = np.abs(self.parameters['nvalue'])
        # Create filter
        centerx = np.ceil(width1 / 2.0) - 1.0
        centery = np.int16(np.ceil(height1 / 2.0) - 1)
        self.row1 = centery - v0
        self.row2 = centery + v0+1
        listx = np.arange(width1)-centerx
        filtershape = 1.0/(1.0 + np.power(listx/u0, 2*n))
        filtershapepad2d = np.zeros((self.row2 - self.row1, filtershape.size))
        filtershapepad2d[:] = np.float64(filtershape)
        self.filtercomplex = filtershapepad2d + filtershapepad2d*1j

        a = pyfftw.empty_aligned((height1, width1), dtype='complex128', n=16)
        b = pyfftw.empty_aligned((height1, width1), dtype='complex128', n=16)
        c = pyfftw.empty_aligned((height1, width1), dtype='complex128', n=16)
        d = pyfftw.empty_aligned((height1, width1), dtype='complex128', n=16)
        self.fft_object = pyfftw.FFTW(a, b, axes=(0, 1))
        self.ifft_object = pyfftw.FFTW(c, d, axes=(0, 1),
                                       direction='FFTW_BACKWARD')

    def process_frames(self, data):
        sino = fft.fftshift(self.fft_object(data[0]))
        sino[self.row1:self.row2] = \
            sino[self.row1:self.row2] * self.filtercomplex
        sino = fft.ifftshift(sino)
        return self.ifft_object(sino).real

    def get_plugin_pattern(self):
        return 'SINOGRAM'

    def get_max_frames(self):
        return 'single'

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The ring artefact removal algorithm used in this processing \
             chain is taken from this work.")
        cite_info.bibtex = \
            ("@article{raven1998numerical,\n" +
             "title={Numerical removal of ring artifacts in \
             microtomography},\n" +
             "author={Raven, Carsten},\n" +
             "journal={Review of scientific instruments},\n" +
             "volume={69},\n" +
             "number={8},\n" +
             "pages={2978--2980},\n" +
             "year={1998},\n" +
             "publisher={AIP Publishing}\n" +
             "}")
        cite_info.endnote = \
            ("%0 Journal Article\n" +
             "%T Numerical removal of ring artifacts in microtomography\n" +
             "%A Raven, Carsten\n" +
             "%J Review of scientific instruments\n" +
             "%V 69\n" +
             "%N 8\n" +
             "%P 2978-2980\n" +
             "%@ 0034-6748\n" +
             "%D 1998\n" +
             "%I AIP Publishing")
        cite_info.doi = "doi: 10.1063/1.1149043"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: dark_flat_field_correction
   :platform: Unix
   :synopsis: A Plugin to apply a simple dark and flatfield correction to raw\
       timeseries data

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import logging
import numpy as np

from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.corrections.base_correction import BaseCorrection
from savu.plugins.utils import register_plugin


@register_plugin
class DarkFlatFieldCorrection(BaseCorrection, CpuPlugin):
    """
    A Plugin to apply a simple dark and flat field correction to data.
    :param pattern: Data processing pattern is 'PROJECTION' or \
        'SINOGRAM'. Default: 'PROJECTION'.
    :param lower_bound: Set all values below the lower_bound to this \
        value. Default: None.
    :param upper_bound: Set all values above the upper bound to this \
        value. Default: None.
    :param warn_proportion: Output a warning if this proportion of values, \
        or greater, are below and/or above the lower/upper bounds, \
        e.g enter 0.05 for 5%. Default: 0.05.
    """

    def __init__(self):
        super(DarkFlatFieldCorrection,
              self).__init__("DarkFlatFieldCorrection")
        self.flag_low_warning = False
        self.flag_high_warning = False
        self.flag = True

    def pre_process(self):
        inData = self.get_in_datasets()[0]
        in_pData = self.get_plugin_in_datasets()[0]
        logging.debug('getting the dark data')
        self.dark = inData.data.dark_mean()
        logging.debug('getting the flat data')
        self.flat = inData.data.flat_mean()
        #np.save('avflats.npy', self.flat)
        #print(np.shape(self.flat))

        pData_shape = in_pData.get_shape()
        tile = [1]*len(pData_shape)
        rot_dim = inData.get_data_dimension_by_axis_label('rotation_angle')
        self.slice_dir = in_pData.get_slice_dimension()

        if self.parameters['pattern'] == 'PROJECTION':
            self._proj_pre_process(inData, pData_shape, tile, rot_dim)
        elif self.parameters['pattern'] == 'SINOGRAM':
            self._sino_pre_process(inData, tile, rot_dim)

        self.flat_minus_dark = self.flat - self.dark
        self.flat_minus_dark[self.flat_minus_dark == 0.0] = 1.0
        self.warn = self.parameters['warn_proportion']
        self.low = self.parameters['lower_bound']
        self.high = self.parameters['upper_bound']
        self.in_pData = in_pData

    def _proj_pre_process(self, data, shape, tile, dim):
        tile[dim] = shape[dim]
        self.convert_size = lambda x: np.tile(x, tile)
        self.process_frames = self.correct_proj

    def _sino_pre_process(self, data, tile, dim):
        pData = data._get_plugin_data()
        full_shape = data.get_shape()
        tile[dim] = full_shape[dim]
        self.process_frames = self.correct_sino
        self.n_plugin_frames = pData.get_shape()[self.slice_dir]

        self.length = full_shape[self.slice_dir]
        self.mfp = pData._get_max_frames_process()
        self.reps_at = int(np.ceil(self.length/float(self.mfp)))

        if len(full_shape) is 3:
            self.convert_size = lambda a, b, x, pad: np.pad(
                    np.tile(x[a:b], tile), pad, 'edge')
        else:
            nSino = \
                full_shape[data.get_data_dimension_by_axis_label('detector_y')]
            self.convert_size = \
                lambda a, b, x, pad: np.pad(
                        np.tile(x[a % nSino:b], tile), pad, 'edge')

    def correct_proj(self, data):
        data = data[0]
        dark = self.convert_size(self.dark)
        flat_minus_dark = self.convert_size(self.flat_minus_dark)
        data = np.nan_to_num((data-dark)/flat_minus_dark)
        self.__data_check(data)
        return data

    def correct_sino(self, data):
        data = data[0]
        sl = self.get_current_slice_list()[0][self.slice_dir]
        count = self.get_process_frames_counter()
        current_idx = self.get_global_frame_index()[count]
        
        start = (current_idx % self.reps_at)*self.mfp
        end = start + len(np.arange(sl.start, sl.stop, sl.step))
        pad = self._get_pad_amount(end)

        dark = self.convert_size(start, end, self.dark, pad)
        flat_minus_dark = \
            self.convert_size(start, end, self.flat_minus_dark, pad)

        data = np.nan_to_num((data-dark)/flat_minus_dark)
        self.__data_check(data)
        return data

    def _get_pad_amount(self, end):
        pad = [[0, 0] for i in range(3)]
        if end > self.length:
            pad[self.slice_dir][1] = end - self.length
        return pad

    def fixed_flag(self):
        if self.parameters['pattern'] == 'PROJECTION':
            return True
        else:
            return False

    def __data_check(self, data):
        # make high and low crop masks and flag if those masks include a large
        # proportion of pixels, as this may indicate a failure
        if not self.low and not self.high:
            return

        if self.low:
            low_crop = data < self.parameters['lower_bound']
            if ((float(low_crop.sum()) / low_crop.size) > self.warn):
                self.flag_low_warning = True
            # Set all cropped values to the crop level
            data[low_crop] = self.parameters['lower_bound']
        if self.high:
            high_crop = data > self.parameters['upper_bound']
            if ((float(high_crop.sum()) / high_crop.size) > self.warn):
                self.flag_high_warning = True
            # Set all cropped values to the crop level
            data[high_crop] = self.parameters['upper_bound']

    def executive_summary(self):
        summary = []
        if self.flag_high_warning:
            summary.append(("WARNING: over %i%% of pixels are being clipped " +
                            "as they have %f times the intensity of the " +
                            "provided flat field. Check your Dark and Flat " +
                            "correction images") % (self.warn*100, self.high))

        if self.flag_low_warning:
            summary.append(("WARNING: over %i%% of pixels are being clipped " +
                            "as they below the expected lower corrected " +
                            "threshold of  %f. Check your Dark and Flat " +
                            "correction images") % (self.warn*100, self.low))
        if len(summary) > 0:
            return summary

        return ["Nothing to Report"]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: testing_iterative_plugin4
   :platform: Unix
   :synopsis: Iterative plugin example
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.iterative_plugin import IterativePlugin


@register_plugin
class TestingIterativePlugin4(BaseFilter, IterativePlugin):
    """
    A plugin to test the iterative plugin driver - switching between sinograms
    and projections on each iteration.
    """

    def __init__(self):
        super(TestingIterativePlugin4, self).\
            __init__("TestingIterativePlugin4")

    def pre_process(self):
        self.set_iterations(3)

    def process_frames(self, data):
        return data[0]

    def post_process(self):
        # option here to break out of the iterations
        # self.set_processing_complete()
        pass

    def setup(self):
        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()

        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')

        out_dataset[0].create_dataset(in_dataset[0])
        # Clone and set as alternating dataset - should I actually do this here?
        self.clone_dataset(out_dataset[1], out_dataset[0])

        # HOw to set more than one pattern associated with a plugin
        out_pData[0].plugin_data_setup('SINOGRAM', 'single') # set first pattern
        out_pData[1].plugin_data_setup('PROJECTION', 'single') # set first pattern

        # Do I need two plugin data objects?  When should I set this?

        # try replacing input dataset with the output dataset *****option to change the pattern on each iteration?
        # option to set any number of patterns one after another
        self.set_iteration_datasets(0, [in_dataset[0]], [out_dataset[0]], pattern='SINOGRAM')  # Add option for pattern to be a dictionary with different pattern for each dataset
        self.set_iteration_datasets(1, [out_dataset[0]], [out_dataset[1]], pattern='PROJECTION') # PROJECTION
        # out_dataset[0] and out_dataset[1] will continue to alternate for the
        # remaining iterations
        self.set_alternating_patterns(['SINOGRAM', 'PROJECTION']) # or explicitly add it as above or a combination of both
        # it could be that the first two iterations have the same pattern and the remainder alternate
        
        # alternatively to above you could just have
#        self.set_iteration_datasets(0, [in_dataset[0]], [out_dataset[0]])
#        self.set_iteration_datasets(1, [out_dataset[0]], [out_dataset[1]])
#        self.set_alternating_patterns(['SINOGRAM', 'PROJECTION']) # what if there is more than on pattern - this should also be a dictionary
        
        # So, to set different patterns there are two ways
        # first way is to add 'pattern' key word to set_iteration_datasets function call
        # second way is to pass a list of patterns to set_alternating_patterns function call
        # in either case, each entry can be a list of a dictionary
        # if a list apply to all datasets
        # if a dictionary, they should be {dataset: pattern} key value pairs
        
        # Now I just have to make this work - can I just create an extra pluginData object for each dataset and alternate between those?
        # Or will I have to calculate mfp/mft every time?
        # Don't forget to call "finalise_datasets" or whatever the function is (usually called after the setup method)

    def nOutput_datasets(self):
        return 2

    def nClone_datasets(self):
        return 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: downsample_filter
   :platform: Unix
   :synopsis: A plugin to filter each frame with a downsample

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import logging
import numpy
import skimage.measure as skim

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class DownsampleFilter(Plugin, CpuPlugin):
    """
    A plugin to reduce the data in the selected dimension by a proportion

    :u*param bin_size: Bin Size for the downsample. Default: 2.
    :u*param mode: One of 'mean', 'median', 'min', 'max'. Default: 'mean'.
    :u*param pattern: One of 'PROJECTION' or 'SINOGRAM'. Default: 'PROJECTION'.
    """

    def __init__(self):
        logging.debug("Starting Downsample Filter")
        super(DownsampleFilter,
              self).__init__("DownsampleFilter")
        self.out_shape = None
        self.mode_dict = { 'mean'  : numpy.mean,
                           'median': numpy.median,
                           'min'   : numpy.min,
                           'max'   : numpy.max }

    def get_output_shape(self, input_data):
        input_shape = input_data.get_shape()
        core_dirs = input_data.get_core_directions()
        output_shape = []
        for i in range(len(input_shape)):
            if i in core_dirs:
                output_shape.append(input_shape[i]/self.parameters['bin_size'])
            else:
                output_shape.append(input_shape[i])

        return output_shape

    def process_frames(self, data):
        logging.debug("Running Downsample data")
        if self.parameters['mode'] in self.mode_dict:
            sampler = self.mode_dict[self.parameters['mode']]
        else:
            logging.warning("Unknown downsample mode. Using 'mean'.")
            sampler = numpy.mean
        block_size = (self.parameters['bin_size'], self.parameters['bin_size'])
        result = skim.block_reduce(data[0], block_size, sampler)
        return result

    def setup(self):
        # get all in and out datasets required by the plugin
        in_dataset, out_dataset = self.get_datasets()
        # get plugin specific instances of these datasets
        in_pData, out_pData = self.get_plugin_datasets()

        plugin_pattern = self.parameters['pattern']
        in_pData[0].plugin_data_setup(plugin_pattern, self.get_max_frames())

        self.out_shape = \
            self.new_shape(in_dataset[0].get_shape(), in_dataset[0])

        out_dataset[0].create_dataset(patterns=in_dataset[0],
                                      axis_labels=in_dataset[0],
                                      shape=self.out_shape)

        out_pData[0].plugin_data_setup(plugin_pattern, self.get_max_frames())

    def new_shape(self, full_shape, data):
        core_dirs = data.get_core_dimensions()
        new_shape = list(full_shape)
        for dim in core_dirs:
            new_shape[dim] = full_shape[dim] // self.parameters['bin_size']
            if (full_shape[dim] % self.parameters['bin_size']) > 0:
                new_shape[dim] += 1
        return tuple(new_shape)

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1

    def get_max_frames(self):
        # This filter processes one frame at a time
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: checkpointing
   :platform: Unix
   :synopsis: A class to organise Savu checkpointing.
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""

import os
import time
import copy
import logging
import numpy as np
from mpi4py import MPI
from shutil import copyfile

import savu.plugins.utils as pu
from savu.data.meta_data import MetaData
from savu.plugins.savers.utils.hdf5_utils import Hdf5Utils


class Checkpointing(object):
    """ Contains all checkpointing associated methods.
    """

    def __init__(self, exp, name='Checkpointing'):
        self._exp = exp
        self._h5 = Hdf5Utils(self._exp)
        self._filename = '_checkpoint.h5'
        self._file = None
        self._start_values = (0, 0, 0)
        self._completed_plugins = 0
        self._level = None
        self._proc_idx = 0
        self._trans_idx = 0
        self._comm = None
        self._timer = None
        self._set_timer()
        self.meta_data = MetaData()

    def _initialise(self, comm):
        """ Create a new checkpoint file """
        with self._h5._open_backing_h5(self._file, 'a', mpi=False) as f:
            self._create_dataset(f, 'transfer_idx', 0)
            self._create_dataset(f, 'process_idx', 0)
            self._create_dataset(
                    f, 'completed_plugins', self._completed_plugins)
        msg = "%s initialise." % self.__class__.__name__
        self._exp._barrier(communicator=comm, msg=msg)

    def _create_dataset(self, f, name, val):
        if name in list(f.keys()):
            f[name][...] = val
        else:
            f.create_dataset(name, data=val, dtype=np.int16)

    def __set_checkpoint_info(self):
        mData = self._exp.meta_data.get
        proc = 'process%d' % mData('process')
        self._folder = os.path.join(mData('out_path'), 'checkpoint')
        self._file = os.path.join(self._folder, proc + self._filename)

        if self._exp.meta_data.get('process') == 0:
            if not os.path.exists(self._folder):
                os.makedirs(os.path.join(self._folder))
        self._exp._barrier(msg='Creating checkpoint folder.')

    def _set_checkpoint_info_from_file(self, level):
        self._level = level
        self.__set_checkpoint_info()
        self.__does_file_exist(self._file, level)

        with self._h5._open_backing_h5(self._file, 'r', mpi=False) as f:
            self._completed_plugins = \
                f['completed_plugins'][...] if 'completed_plugins' in f else 0
            self._proc_idx = f['process_idx'][...] if 'process_idx' in f and \
                level == 'subplugin' else 0
            self._trans_idx = f['transfer_idx'][...] if 'transfer_idx' in f \
                and level == 'subplugin' else 0
            # for testing
            self.__set_start_values(
                    self._completed_plugins, self._trans_idx, self._proc_idx)
            self.__set_dataset_metadata(f, 'in_data')
            self.__set_dataset_metadata(f, 'out_data')

        self.__load_data()
        msg = "%s _set_checkpoint_info_from_file" % self.__class__.__name__
        self._exp._barrier(msg=msg)

    def __does_file_exist(self, thefile, level):
        if not os.path.exists(thefile):
            if level == 'plugin':
                proc0 = os.path.join(self._folder, 'process0' + self._filename)
                self.__does_file_exist(proc0, None)
                copyfile(proc0, self._file)
                return
            raise Exception("No checkpoint file found.")

    def __set_dataset_metadata(self, f, dtype):
        self.meta_data.set(dtype, {})
        if dtype not in list(f.keys()):
            return
        entry = f[dtype]
        for name, gp in entry.items():
            data_entry = gp.require_group('meta_data')
            for key, value in data_entry.items():
                self.meta_data.set([dtype, name, key], value[key][...])

    def _get_dataset_metadata(self, dtype, name):
        return self._data_meta_data(dtype)

    def set_completed_plugins(self, n):
        self._completed_plugins = n

    def __load_data(self):
        self._exp.meta_data.set('checkpoint_loader', True)
        temp = self._exp.meta_data.get('data_file')
        nxsfile = self._exp.meta_data.get('nxs_filename')
        self._exp.meta_data.set('data_file', nxsfile)
        pid = 'savu.plugins.loaders.savu_nexus_loader'
        pu.plugin_loader(self._exp, {'id': pid, 'data': {}})
        self._exp.meta_data.delete('checkpoint_loader')
        self._exp.meta_data.set('data_file', temp)

    def output_plugin_checkpoint(self):
        self._completed_plugins += 1
        self.__write_plugin_checkpoint()
        self._reset_indices()

    def get_checkpoint_plugin(self):
        checkpoint_flag = self._exp.meta_data.get('checkpoint')
        if not checkpoint_flag:
            self.__set_checkpoint_info()
            self._initialise(MPI.COMM_WORLD)
        else:
            self._set_checkpoint_info_from_file(checkpoint_flag)
        return self._completed_plugins

    def is_time_to_checkpoint(self, transport, ti, pi):
        interval = self._exp.meta_data.get(
                ['system_params', 'checkpoint_interval'])
        end = time.time()
        if (end - self._get_timer()) > interval:
            self.__write_subplugin_checkpoint(ti, pi)
            self._set_timer()
            transport._transport_checkpoint()
            return transport._transport_kill_signal()
        return False

    def _get_checkpoint_params(self):
        return self._level, self._completed_plugins

    def __write_subplugin_checkpoint(self, ti, pi):
        with self._h5._open_backing_h5(self._file, 'a', mpi=False) as f:
            f['transfer_idx'][...] = ti
            f['process_idx'][...] = pi

    def __write_plugin_checkpoint(self):
        with self._h5._open_backing_h5(self._file, 'a', mpi=False) as f:
            f['completed_plugins'][...] = self._completed_plugins
            f['transfer_idx'][...] = 0
            f['process_idx'][...] = 0

    def _reset_indices(self):
        self._trans_idx = 0
        self._proc_idx = 0

    def get_trans_idx(self):
        return self._trans_idx

    def get_proc_idx(self):
        return self._proc_idx

    def get_level(self):
        return self._level

    def _set_timer(self):
        self._timer = time.time()

    def _get_timer(self):
        return self._timer

    def __set_start_values(self, v1, v2, v3):
        self._start_values = (copy.copy(v1), copy.copy(v2), copy.copy(v3))

    def get_start_values(self):
        return self._start_values



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: min_and_max
   :platform: Unix
   :synopsis: A plugin to calculate the min and max of each frame
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""
import logging
import numpy as np

from scipy.ndimage import gaussian_filter
from savu.plugins.plugin import Plugin
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
import savu.core.utils as cu


@register_plugin
class MinAndMax(Plugin, CpuPlugin):
    """
    A plugin to calculate the min and max values of each slice (as determined \
    by the pattern parameter)

    :u*param pattern: How to slice the data. Default: 'VOLUME_XZ'.
    :param smoothing: Apply a smoothing filter or not. Default: True.
    :u*param masking: Apply a circular mask or not. Default: True.
    :u*param ratio: Used to calculate the circular mask. If not provided, \
        it is calculated using the center of rotation. Default: None.
    :u*param method: Method to find the global min and the global max. \
        Available options: 'extrema', 'percentile'. Default: 'percentile'.
    :u*param p_range: Percentage range if use the 'percentile' method. \
        Default: [0.0, 100.0].
    :param out_datasets: The default names. Default: ['the_min','the_max'].
    """

    def __init__(self):
        super(MinAndMax, self).__init__("MinAndMax")

    def circle_mask(self, width, ratio):
        # Create a circle mask.
        mask = np.zeros((width, width), dtype=np.float32)
        center = width // 2
        radius = ratio * center
        y, x = np.ogrid[-center:width - center, -center:width - center]
        mask_check = x * x + y * y <= radius * radius
        mask[mask_check] = 1.0
        return mask

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()[0]
        in_meta_data = self.get_in_meta_data()[0]
        data = self.get_in_datasets()[0]
        data_shape = data.get_shape()
        width = data_shape[-1]
        use_mask = self.parameters['masking']
        if use_mask is True:
            ratio = self.parameters['ratio']
            if ratio is None:
                try:
                    cor = np.min(in_meta_data.get('centre_of_rotation'))
                    ratio = (min(cor, abs(width - cor))) / (width * 0.5)
                except KeyError:
                    ratio = 1.0
            self.mask = self.circle_mask(width, ratio)
        else:
            self.mask = np.ones((width, width), dtype=np.float32)
        self.method = self.parameters['method']
        if not (self.method == 'percentile' or self.method == 'extrema'):
            msg = "\n***********************************************\n"\
                "!!! ERROR !!! -> Wrong method. Please use only one of "\
                "the provided options \n"\
                "***********************************************\n"
            logging.warning(msg)
            cu.user_message(msg)
            raise ValueError(msg)
        self.p_min, self.p_max = np.sort(np.clip(np.asarray(
            self.parameters['p_range'], dtype=np.float32), 0.0, 100.0))

    def process_frames(self, data):
        use_filter = self.parameters['smoothing']
        frame = np.nan_to_num(data[0])
        if use_filter is True:
            frame = gaussian_filter(frame, (3, 3)) * self.mask
        else:
            frame = frame * self.mask
        if self.method == 'percentile':
            list_out = [np.array(
                [np.percentile(frame, self.p_min)], dtype=np.float32),
                np.array([np.percentile(frame, self.p_max)], dtype=np.float32)]
        else:
            list_out = [np.array([np.min(frame)], dtype=np.float32),
                        np.array([np.max(frame)], dtype=np.float32)]
        return list_out

    def post_process(self):
        in_datasets, out_datasets = self.get_datasets()
        the_min = np.squeeze(out_datasets[0].data[...])
        the_max = np.squeeze(out_datasets[1].data[...])
        pattern = self._get_pattern()
        in_datasets[0].meta_data.set(['stats', 'min', pattern], the_min)
        in_datasets[0].meta_data.set(['stats', 'max', pattern], the_max)

    def setup(self):
        in_dataset, out_datasets = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup(self._get_pattern(), 'single')

        slice_dirs = list(in_dataset[0].get_slice_dimensions())
        orig_shape = in_dataset[0].get_shape()
        new_shape = (np.prod(np.array(orig_shape)[slice_dirs]), 1)

        labels = ['x.pixels', 'y.pixels']
        for i in range(len(out_datasets)):
            out_datasets[i].create_dataset(shape=new_shape, axis_labels=labels,
                                           remove=True, transport='hdf5')
            out_datasets[i].add_pattern(
                "METADATA", core_dims=(1,), slice_dims=(0,))
            out_pData[i].plugin_data_setup('METADATA', 'single')

    def _get_pattern(self):
        return self.parameters['pattern']

    def nOutput_datasets(self):
        return 2



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: ptypy_compact
   :platform: Unix
   :synopsis: A plugin to perform ptychography using ptypy

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.utils import register_plugin
from savu.plugins.ptychography.base_ptycho import BasePtycho
import numpy as np
from ptypy.core import Ptycho
from ptypy import utils as u


@register_plugin
class PtypyCompact(BasePtycho):
    """
    This plugin performs ptychography using the ptypy package.
    :param data_center: See online documentation. Default: None.
    :param data_orientation:See online documentation . Default: None.
    :param data_auto_center: See online documentation. Default: True.
    :param geometry_shape:See online documentation . Default: 128.
    :param illumination_photons: See online documentation. Default: None.
    :param illumination_diversity_noise: See online documentation. Default: None.
    :param illumination_diversity_power: See online documentation. Default: 0.1.
    :param illumination_diversity_shift: See online documentation. Default: None.
    :param sample_model: See online documentation. Default: None.
    :param sample_fill: See online documentation. Default: 1.
    :param sample_diversity_noise: See online documentation. Default: None.
    :param sample_diversity_power: See online documentation. Default: 0.1.
    :param sample_diversity_shift: See online documentation. Default: None.
    :param coherence_num_probe_modes: See online documentation. Default: 1.
    :param coherence_num_object_modes: See online documentation. Default: 1.
    :param coherence_spectrum: See online documentation. Default: None.
    :param coherence_object_dispersion: See online documentation. Default: None.
    :param coherence_probe_dispersion: See online documentation. Default: None.
    :param common_numiter_contiguous: See online documentation. Default: 1.
    :param common_probe_support: See online documentation. Default: 0.7.
    :param common_clip_object: See online documentation. Default: None.
    :param DM_num_iter: See online documentation. Default: 30.
    :param DM_alpha: See online documentation. Default: 1.
    :param DM_probe_update_start: See online documentation. Default: 2.
    :param DM_update_object_first: See online documentation. Default: True.
    :param DM_overlap_converge_factor: See online documentation. Default: 0.05.
    :param DM_overlap_max_iterations: See online documentation. Default: 10.
    :param DM_probe_inertia: See online documentation. Default: 0.001.
    :param DM_object_inertia: See online documentation. Default: 0.1.
    :param DM_fourier_relax_factor: See online documentation. Default: 0.01.
    :param DM_obj_smooth_std: See online documentation. Default: 20.
    :param ML_num_iter: See online documentation. Default: 5.
    :param ML_type: See online documentation. Default: "gaussian".
    :param ML_floating_intensities: See online documentation. Default: False.
    :param ML_intensity_renormalization: See online documentation. Default: 1.
    :param ML_reg_del2: See online documentation. Default: True.
    :param ML_reg_del2_amplitude: See online documentation. Default: 0.01.
    :param ML_smooth_gradient: See online documentation. Default: 0.
    :param ML_scale_precond: See online documentation. Default: False.
    :param ML_scale_probe_object: See online documentation. Default: 1.
    :param ML_probe_update_start: See online documentation. Default: 0.

    """

    def __init__(self):
        super(PtypyCompact, self).__init__("PtypyCompact")

    def pre_process(self):
        in_dataset, out_datasets = self.get_datasets()
        in_d1 = in_dataset[0]
        sh = in_d1.get_shape()[-1]
        
        p, r = self.parse_params()
        ###
        self.get
        self.p = p
        self.r = r

    def process_frames(self, data):
        self.r.data = data[0]
        self.p.scans.savu.data.recipe = self.r
        # reconstruct
        P = Ptycho(self.p,level=5)
        # now post process
        object_stack = P.obj.storages['S00G00'].data.astype(np.float)
        probe_stack = P.probe.storages['S00G00'].data.astype(np.float)
        positions = self.get_positions()
        return [probe_stack, object_stack, positions]#] add fourier error, realspace error

    def parse_params(self):
        p = u.Param()
        p.scan = u.Param()
        p.scan.data = u.Param()
        p.scan.data.center = self.parameters['data_center']
        p.scan.data.orientation = self.parameters['data_orientation']
        p.scan.data.auto_center = self.parameters['data_auto_center']
        p.scan.illumination = u.Param()
        p.scan.illumination.photons = self.parameters['illumination_photons']
        p.scan.illumination.diversity = u.Param()
        p.scan.illumination.diversity.noise = self.parameters['illumination_diversity_noise']
        p.scan.illumination.diversity.power = self.parameters['illumination_diversity_power']
        p.scan.illumination.diversity.shift = self.parameters['illumination_diversity_shift']
        p.scan.sample = u.Param()
        p.scan.sample.model = self.parameters['sample_model']
        p.scan.sample.fill = self.parameters['sample_fill']
        p.scan.sample.recon = u.Param()
        p.scan.sample.diversity = u.Param()
        p.scan.sample.diversity.noise = self.parameters['sample_diversity_noise']
        p.scan.sample.diversity.power = self.parameters['sample_diversity_power']
        p.scan.sample.diversity.shift = self.parameters['sample_diversity_shift']
        p.scan.coherence = u.Param()
        p.scan.coherence.num_probe_modes = self.parameters['coherence_num_probe_modes']
        p.scan.coherence.num_object_modes = self.parameters['coherence_num_object_modes']
        p.scan.coherence.spectrum = self.parameters['coherence_spectrum']
        p.scan.coherence.object_dispersion = self.parameters['coherence_object_dispersion']
        p.scan.coherence.probe_dispersion = self.parameters['coherence_probe_dispersion']
        p.scan.geometry = u.Param()
        p.scan.geometry.shape = self.parameters['geometry_shape']
        p.engine = u.Param()
        p.engine.common = u.Param()
        p.engine.common.probe_support = self.parameters['common_probe_support']
        p.engine.common.clip_object = self.parameters['common_clip_object'] #     turn off all of the output things
        p.io = u.Param()
        p.io.home = "./" ## (05) Base directory for all I/O
        p.io.rfile = False ## (06) Reconstruction file name (or format string)
        p.io.autosave = u.Param()
        p.io.autosave.active = False ## (08) Activation switch
        p.io.autosave.interval = -1 ## (09) Auto-save interval
    # p.io.autosave.rfile = "dumps/%(run)s/%(run)s_%(engine)s_%(iterations)04d.ptyr" ## (10) Auto-save file name (or format string)
    ## (11) Server / Client parameters
        p.io.interaction = u.Param()
        p.io.interaction.active = True
        p.io.autoplot = u.Param()
        p.io.autoplot.imfile = "./" ## (17) Plot images file name (or format string)
        p.io.autoplot.interval = 5 ## (18) Number of iterations between plot updates
        p.io.autoplot.threaded = True ## (19) Live plotting switch
        p.io.autoplot.layout = u.Param() ## (20) Options for default plotter or template name
        p.io.autoplot.dump = False ## (21) Switch to dump plots as image files
        p.io.autoplot.make_movie = False
        if self.parameters['DM_num_iter'] is not None:
            p.engine.DM = u.Param()
            p.engine.DM.name = "DM"
            p.engine.DM.alpha = self.parameters['DM_alpha']
            p.engine.DM.probe_update_start = self.parameters['DM_probe_update_start']
            p.engine.DM.update_object_first = self.parameters['DM_update_object_first']
            p.engine.DM.overlap_converge_factor = self.parameters['DM_overlap_converge_factor']
            p.engine.DM.overlap_max_iterations = self.parameters['DM_overlap_max_iterations']
            p.engine.DM.probe_inertia = self.parameters['DM_probe_inertia']
            p.engine.DM.object_inertia = self.parameters['DM_object_inertia']
            p.engine.DM.fourier_relax_factor = self.parameters['DM_fourier_relax_factor']
            p.engine.DM.obj_smooth_std = self.parameters['DM_obj_smooth_std']
            p.engines = u.Param()
            p.engines.engine_00 = u.Param()
            p.engines.engine_00.name = 'DM'
            p.engines.engine_00.numiter = self.parameters['DM_num_iter']
        if self.parameters['ML_num_iter'] is not None:
            p.engine.ML = u.Param()
            p.engine.ML.name = "ML"
            p.engine.ML.type = self.parameters['ML_type']
            p.engine.ML.floating_intensities = self.parameters['ML_floating_intensities']
            p.engine.ML.intensity_renormalization = self.parameters['ML_intensity_renormalization']
            p.engine.ML.reg_del2 = self.parameters['ML_reg_del2']
            p.engine.ML.reg_del2_amplitude = self.parameters['ML_reg_del2_amplitude']
            p.engine.ML.smooth_gradient = self.parameters['ML_smooth_gradient']
            p.engine.ML.scale_precond = self.parameters['ML_scale_precond']
            p.engine.ML.scale_probe_object = self.parameters['ML_scale_probe_object']
            p.engine.ML.probe_update_start = self.parameters['ML_probe_update_start'] #             p.engines = u.Param()
            p.engines.engine_01 = u.Param()
            p.engines.engine_01.name = 'ML'
            p.engines.engine_01.numiter = self.parameters['ML_num_iter']
        p.scans = u.Param()
        p.scans.savu = u.Param()
        p.scans.savu.if_conflict_use_meta = True
        p.scans.savu.data = u.Param()
        p.scans.savu.data.source = 'savu'
        p.scans.savu.data.dfile = None
        p.scans.savu.data.shape = self.parameters['geometry_shape']
        p.scans.savu.data.save = None
        p.ipython_kernel = False
        r = u.Param()
        r.positions = self.get_positions().T ### this stuff can come out of the nexus file
        p.scans.savu.data.psize = 55e-6
        p.scans.savu.data.distance = 1.59
        p.scans.savu.data.orientation = 2
        p.scans.savu.data.energy = 9.1
        p.scan.illumination = u.Param()
        p.scan.illumination.model = None
        p.scan.illumination.photons = 1e8
        p.scan.illumination.aperture = u.Param()
        p.scan.illumination.aperture.form = "circ"
        p.scan.illumination.aperture.size = 400e-6
        p.scan.illumination.propagation = u.Param()
        p.scan.illumination.propagation.focussed = 440.5e-3
        p.scan.illumination.propagation.parallel = 2.3e-3
        r.mask = np.ones((p.scans.savu.data.shape, p.scans.savu.data.shape))
#         probe_size = tuple(u.expect2(p.scan.geometry.shape)) + (p.scan.coherence.num_probe_modes, )
        
#         self.set_size_probe(probe_size)
        return p, r

    def get_num_probe_modes(self):
        return self.p.scan.coherence.num_probe_modes

    def set_size_object(self, in_d1, positions, pobj):
        positions = self.get_positions()
        p, r = self.parse_params()
        sh = p.scans.savu.data.shape
        r.data = np.ones((positions.shape[-1],sh,sh))
        p.scans.savu.data.recipe = r
        P = Ptycho(p,level=2)
        object_shape = P.obj.storages['S00G00'].data[0].shape
        self.obj_shape = object_shape + (self.get_num_object_modes(),)
        #print "object shape is" + str(self.obj_shape)
    
    def set_size_probe(self, probe_shape):
        self.p, self.r = self.parse_params()
        sh = self.p.scans.savu.data.shape
        self.probe_size = (1,)+tuple(u.expect2(sh)) + (self.get_num_probe_modes(),)
        #print "probe size is" + str(self.probe_size)




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin
   :platform: Unix
   :synopsis: Base class for all plugins used by Savu

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import ast
import copy
import logging
import inspect
import numpy as np

import savu.plugins.docstring_parser as doc
from savu.plugins.plugin_datasets import PluginDatasets


class Plugin(PluginDatasets):
    """
    The base class from which all plugins should inherit.
    :param in_datasets: Create a list of the dataset(s) to \
        process. Default: [].
    :param out_datasets: Create a list of the dataset(s) to \
        create. Default: [].
    """

    def __init__(self, name="Plugin"):
        super(Plugin, self).__init__(name)
        self.name = name
        self.parameters = {}
        self.parameters_types = {}
        self.parameters_desc = {}
        self.parameters_hide = []
        self.parameters_user = []
        self.chunk = False
        self.docstring_info = {}
        self.slice_list = None
        self.global_index = None
        self.pcount = 0
        self.exp = None
        self.check = False

    def initialise(self, params, exp, check=False):
        self.check = check
        self.exp = exp
        self._set_parameters(copy.deepcopy(params))
        self._main_setup()

    def _main_setup(self):
        """ Performs all the required plugin setup.

        It sets the experiment, then the parameters and replaces the
        in/out_dataset strings in ``self.parameters`` with the relevant data
        objects. It then creates PluginData objects for each of these datasets.
        """
        self._set_plugin_datasets()
        self._reset_process_frames_counter()
        self.setup()
        self.set_filter_padding(*(self.get_plugin_datasets()))
        self._finalise_plugin_datasets()
        self._finalise_datasets()

    def _reset_process_frames_counter(self):
        self.pcount = 0

    def get_process_frames_counter(self):
        return self.pcount

    def _set_parameters_this_instance(self, indices):
        """ Determines the parameters for this instance of the plugin, in the
        case of parameter tuning.

        param np.ndarray indices: the index of the current value in the
            parameter tuning list.
        """
        dims = set(self.multi_params_dict.keys())
        count = 0
        for dim in dims:
            info = self.multi_params_dict[dim]
            name = info['label'].split('_param')[0]
            self.parameters[name] = info['values'][indices[count]]
            count += 1

    def set_filter_padding(self, in_data, out_data):
        """
        Should be overridden to define how wide the frame should be for each
        input data set
        """
        return {}

    def setup(self):
        """
        This method is first to be called after the plugin has been created.
        It determines input/output datasets and plugin specific dataset
        information such as the pattern (e.g. sinogram/projection).
        """
        logging.error("set_up needs to be implemented")
        raise NotImplementedError("setup needs to be implemented")

    def _populate_default_parameters(self):
        """
        This method should populate all the required parameters with default
        values.  it is used for checking to see if new parameter values are
        appropriate

        It makes use of the classes including parameter information in the
        class docstring such as this

        :param error_threshold: Convergence threshold. Default: 0.001.
        """
        hidden_items = []
        user_items = []
        params = []
        not_params = []
        for clazz in inspect.getmro(self.__class__)[::-1]:
            if clazz != object:
                desc = doc.find_args(clazz, self)
                self.docstring_info['warn'] = desc['warn']
                self.docstring_info['info'] = desc['info']
                self.docstring_info['synopsis'] = desc['synopsis']
                params.extend(desc['param'])
                if desc['hide_param']:
                    hidden_items.extend(desc['hide_param'])
                if desc['user_param']:
                    user_items.extend(desc['user_param'])
                if desc['not_param']:
                    not_params.extend(desc['not_param'])
        self._add_item(params, not_params)
        user_items = [u for u in user_items if u not in not_params]
        hidden_items = [h for h in hidden_items if h not in not_params]
        user_items = list(set(user_items).difference(set(hidden_items)))
        self.parameters_hide = hidden_items
        self.parameters_user = user_items
        self.final_parameter_updates()

    def _add_item(self, item_list, not_list):
        true_list = [i for i in item_list if i['name'] not in not_list]
        for item in true_list:
            self.parameters[item['name']] = item['default']
            self.parameters_types[item['name']] = item['dtype']
            self.parameters_desc[item['name']] = item['desc']

    def delete_parameter_entry(self, param):
        if param in list(self.parameters.keys()):
            del self.parameters[param]
            del self.parameters_types[param]
            del self.parameters_desc[param]

    def initialise_parameters(self):
        self.parameters = {}
        self.parameters_types = {}
        self._populate_default_parameters()
        self.multi_params_dict = {}
        self.extra_dims = []

    def _set_parameters(self, parameters):
        """
        This method is called after the plugin has been created by the
        pipeline framework.  It replaces ``self.parameters`` default values
        with those given in the input process list.

        :param dict parameters: A dictionary of the parameters for this \
        plugin, or None if no customisation is required.
        """
        self.initialise_parameters()
        # reverse sorting added on Python 3 conversion to make the behaviour
        # similar (hopefully the same) as on Python 2
        for key in parameters.keys():
            if key in self.parameters.keys():
                value = self.__convert_multi_params(parameters[key], key)
                self.parameters[key] = value
            else:
                error = ("Parameter '%s' is not valid for plugin %s. \nTry "
                         "opening and re-saving the process list in the "
                         "configurator to auto remove \nobsolete parameters."
                         % (key, self.name))
                raise ValueError(error)

    def __convert_multi_params(self, value, key):
        """ Set up parameter tuning.

        Convert parameter value to a list if it uses parameter tuning and set
        associated parameters, so the framework knows the new size of the data
        and which plugins to re-run.
        """
        dtype = self.parameters_types[key]
        if isinstance(value, str) and ';' in value:
            value = value.split(';')
            if ":" in value[0]:
                seq = value[0].split(':')
                seq = [eval(s) for s in seq]
                value = list(np.arange(seq[0], seq[1], seq[2]))
                if len(value) == 0:
                    raise RuntimeError(
                        'No values for tuned parameter "{}", '
                        'ensure start:stop:step; values are valid.'.format(key))
            if not isinstance(value[0], dtype):
                try:
                    value.remove('')
                except Exception:
                    pass
                if isinstance(value[0], str):
                    value = [ast.literal_eval(i) for i in value]
                value = list(map(dtype, value))
            label = key + '_params.' + type(value[0]).__name__
            self.multi_params_dict[len(self.multi_params_dict)] = \
                {'label': label, 'values': value}
            self.extra_dims.append(len(value))
        return value

    def get_parameters(self, name):
        """ Return a plugin parameter

        :params str name: parameter name (dictionary key)
        :returns: the associated value in ``self.parameters``
        :rtype: dict value
        """
        return self.parameters[name]

    def base_pre_process(self):
        """ This method is called after the plugin has been created by the
        pipeline framework as a pre-processing step.
        """
        pass

    def pre_process(self):
        """ This method is called immediately after base_pre_process(). """
        pass

    def base_process_frames_before(self, data):
        """ This method is called before each call to process frames """
        return data

    def base_process_frames_after(self, data):
        """ This method is called directly after each call to process frames \
        and before returning the data to file."""
        return data

    def plugin_process_frames(self, data):
        frames = self.base_process_frames_after(self.process_frames(
                self.base_process_frames_before(data)))
        self.pcount += 1
        return frames

    def process_frames(self, data):
        """
        This method is called after the plugin has been created by the
        pipeline framework and forms the main processing step

        :param data: A list of numpy arrays for each input dataset.
        :type data: list(np.array)
        """

        logging.error("process frames needs to be implemented")
        raise NotImplementedError("process needs to be implemented")

    def post_process(self):
        """
        This method is called after the process function in the pipeline
        framework as a post-processing step. All processes will have finished
        performing the main processing at this stage.

        :param exp: An experiment object, holding input and output datasets
        :type exp: experiment class instance
        """
        pass

    def base_post_process(self):
        """ This method is called immediately after post_process(). """
        pass

    def set_preview(self, data, params):
        if not params:
            return True
        preview = data.get_preview()
        orig_indices = preview.get_starts_stops_steps()
        nDims = len(orig_indices[0])
        no_preview = [[0]*nDims, data.get_shape(), [1]*nDims, [1]*nDims]

        # Set previewing params if previewing has not already been applied to
        # the dataset.
        if no_preview == orig_indices:
            data.get_preview().revert_shape = data.get_shape()
            data.get_preview().set_preview(params)
            return True
        return False

    def _clean_up(self):
        """ Perform necessary plugin clean up after the plugin has completed.
        """
        self._clone_datasets()
        self.__copy_meta_data()
        self.__set_previous_patterns()
        self.__clean_up_plugin_data()

    def __copy_meta_data(self):
        """
        Copy all metadata from input datasets to output datasets, except axis
        data that is no longer valid.
        """
        remove_keys = self.__remove_axis_data()
        in_meta_data, out_meta_data = self.get()
        copy_dict = {}
        for mData in in_meta_data:
            temp = copy.deepcopy(mData.get_dictionary())
            copy_dict.update(temp)

        for i in range(len(out_meta_data)):
            temp = copy_dict.copy()
            for key in remove_keys[i]:
                if temp.get(key, None) is not None:
                    del temp[key]
            temp.update(out_meta_data[i].get_dictionary())
            out_meta_data[i]._set_dictionary(temp)

    def __set_previous_patterns(self):
        for data in self.get_out_datasets():
            data._set_previous_pattern(
                copy.deepcopy(data._get_plugin_data().get_pattern()))

    def __remove_axis_data(self):
        """
        Returns a list of meta_data entries corresponding to axis labels that
        are not copied over to the output datasets
        """
        in_datasets, out_datasets = self.get_datasets()
        all_in_labels = []
        for data in in_datasets:
            axis_keys = data.get_axis_label_keys()
            all_in_labels = all_in_labels + axis_keys

        remove_keys = []
        for data in out_datasets:
            axis_keys = data.get_axis_label_keys()
            remove_keys.append(set(all_in_labels).difference(set(axis_keys)))

        return remove_keys

    def __clean_up_plugin_data(self):
        """ Remove pluginData object encapsulated in a dataset after plugin
        completion.
        """
        in_data, out_data = self.get_datasets()
        data_object_list = in_data + out_data
        for data in data_object_list:
            data._clear_plugin_data()

    def _revert_preview(self, in_data):
        """ Revert dataset back to original shape if previewing was used in a
        plugin to reduce the data shape but the original data shape should be
        used thereafter. Remove previewing if it was added in the plugin.
        """
        for data in in_data:
            if data.get_preview().revert_shape:
                data.get_preview()._unset_preview()

    def set_global_frame_index(self, frame_idx):
        self.global_index = frame_idx

    def get_global_frame_index(self):
        """ Get the global frame index. """
        return self.global_index


    def set_current_slice_list(self, sl):
        self.slice_list = sl

    def get_current_slice_list(self):
        """ Get the slice list of the current frame being processed. """
        return self.slice_list

    def get_slice_dir_reps(self, nData):
        """ Return the periodicity of the main slice direction.

        :params int nData: The number of the dataset in the list.
        """
        slice_dir = \
            self.get_plugin_in_datasets()[nData].get_slice_directions()[0]
        sl = [sl[slice_dir] for sl in self.slice_list]
        reps = [i for i in range(len(sl)) if sl[i] == sl[0]]
        return np.diff(reps)[0] if len(reps) > 1 else 1

    def nInput_datasets(self):
        """
        The number of datasets required as input to the plugin

        :returns:  Number of input datasets

        """
        return 1

    def nOutput_datasets(self):
        """
        The number of datasets created by the plugin

        :returns:  Number of output datasets

        """
        return 1

    def nClone_datasets(self):
        """ The number of output datasets that have an clone - i.e. they take\
        it in turns to be used as output in an iterative plugin.
        """
        return 0

    def nFrames(self):
        """ The number of frames to process during each call to process_frames.
        """
        return 'single'

    def final_parameter_updates(self):
        """ An opportunity to update the parameters after they have been set.
        """
        pass

    def get_citation_information(self):
        """
        Gets the Citation Information for a plugin

        :returns:  A populated savu.data.plugin_info.CitationInfomration

        """
        return None

    def executive_summary(self):
        """ Provide a summary to the user for the result of the plugin.

        e.g.
         - Warning, the sample may have shifted during data collection
         - Filter operated normally

        :returns:  A list of string summaries
        """
        return ["Nothing to Report"]



# Copyright 201i Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_data
   :platform: Unix
   :synopsis: Contains the PluginData class. Each Data set used in a plugin \
       has a PluginData object encapsulated within it, for the duration of a \
       plugin run.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
import sys
import copy
import h5py
import logging
import numpy as np
from fractions import gcd

from savu.data.meta_data import MetaData
from savu.data.data_structures.data_add_ons import Padding


class PluginData(object):
    """ The PluginData class contains plugin specific information about a Data
    object for the duration of a plugin.  An instance of the class is
    encapsulated inside the Data object during the plugin run
    """

    def __init__(self, data_obj, plugin=None):
        self.data_obj = data_obj
        self._preview = None
        self.data_obj._set_plugin_data(self)
        self.meta_data = MetaData()
        self.padding = None
        self.pad_dict = None
        self.shape = None
        self.core_shape = None
        self.multi_params = {}
        self.extra_dims = []
        self._plugin = plugin
        self.fixed_dims = True
        self.split = None
        self.boundary_padding = None
        self.no_squeeze = False
        self.pre_tuning_shape = None
        self._frame_limit = None
        self._increase_rank = 0

    def _get_preview(self):
        return self._preview

    def get_total_frames(self):
        """ Get the total number of frames to process (all MPI processes).

        :returns: Number of frames
        :rtype: int
        """
        temp = 1
        slice_dir = \
            self.data_obj.get_data_patterns()[
                self.get_pattern_name()]["slice_dims"]
        for tslice in slice_dir:
            temp *= self.data_obj.get_shape()[tslice]
        return temp

    def __set_pattern(self, name, first_sdim=None):
        """ Set the pattern related information in the meta data dict.
        """
        pattern = self.data_obj.get_data_patterns()[name]
        self.meta_data.set("name", name)
        self.meta_data.set("core_dims", pattern['core_dims'])
        self.__set_slice_dimensions(first_sdim=first_sdim)

    def get_pattern_name(self):
        """ Get the pattern name.

        :returns: the pattern name
        :rtype: str
        """
        try:
            name = self.meta_data.get("name")
            return name
        except KeyError:
            raise Exception("The pattern name has not been set.")

    def get_pattern(self):
        """ Get the current pattern.

        :returns: dict of the pattern name against the pattern.
        :rtype: dict
        """
        pattern_name = self.get_pattern_name()
        return {pattern_name: self.data_obj.get_data_patterns()[pattern_name]}

    def __set_shape(self):
        """ Set the shape of the plugin data processing chunk.
        """
        core_dir = self.data_obj.get_core_dimensions()
        slice_dir = self.data_obj.get_slice_dimensions()
        dirs = list(set(core_dir + (slice_dir[0],)))
        slice_idx = dirs.index(slice_dir[0])
        dshape = self.data_obj.get_shape()
        shape = []
        for core in set(core_dir):
            shape.append(dshape[core])
        self.__set_core_shape(tuple(shape))

        mfp = self._get_max_frames_process()
        if mfp > 1 or self._get_no_squeeze():
            shape.insert(slice_idx, mfp)
        self.shape = tuple(shape)

    def _set_shape_transfer(self, slice_size):
        dshape = self.data_obj.get_shape()
        shape_before_tuning = self._get_shape_before_tuning()
        add = [1]*(len(dshape) - len(shape_before_tuning))
        slice_size = slice_size + add

        core_dir = self.data_obj.get_core_dimensions()
        slice_dir = self.data_obj.get_slice_dimensions()
        shape = [None]*len(dshape)
        for dim in core_dir:
            shape[dim] = dshape[dim]
        i = 0
        for dim in slice_dir:
            shape[dim] = slice_size[i]
            i += 1
        return tuple(shape)

    def __get_slice_size(self, mft):
        """ Calculate the number of frames transfer in each dimension given
            mft. """
        dshape = list(self.data_obj.get_shape())

        if 'fixed_dimensions' in list(self.meta_data.get_dictionary().keys()):
            fixed_dims = self.meta_data.get('fixed_dimensions')
            for d in fixed_dims:
                dshape[d] = 1

        dshape = [dshape[i] for i in self.meta_data.get('slice_dims')]
        size_list = [1]*len(dshape)
        i = 0

        while(mft > 1 and i < len(size_list)):
            size_list[i] = min(dshape[i], mft)
            mft //= np.prod(size_list) if np.prod(size_list) > 1 else 1
            i += 1

        self.meta_data.set('size_list', size_list)
        return size_list

    def set_bytes_per_frame(self):
        """ Return the size of a single frame in bytes. """
        nBytes = self.data_obj.get_itemsize()
        dims = list(self.get_pattern().values())[0]['core_dims']
        frame_shape = [self.data_obj.get_shape()[d] for d in dims]
        b_per_f = np.prod(frame_shape)*nBytes
        return frame_shape, b_per_f

    def get_shape(self):
        """ Get the shape of the data (without padding) that is passed to the
        plugin process_frames method.
        """
        return self.shape

    def _set_padded_shape(self):
        pass

    def get_padded_shape(self):
        """ Get the shape of the data (with padding) that is passed to the
        plugin process_frames method.
        """
        return self.shape

    def get_shape_transfer(self):
        """ Get the shape of the plugin data to be transferred each time.
        """
        return self.meta_data.get('transfer_shape')

    def __set_core_shape(self, shape):
        """ Set the core shape to hold only the shape of the core dimensions
        """
        self.core_shape = shape

    def get_core_shape(self):
        """ Get the shape of the core dimensions only.

        :returns: shape of core dimensions
        :rtype: tuple
        """
        return self.core_shape

    def _set_shape_before_tuning(self, shape):
        """ Set the shape of the full dataset used during each run of the \
        plugin (i.e. ignore extra dimensions due to parameter tuning). """
        self.pre_tuning_shape = shape

    def _get_shape_before_tuning(self):
        """ Return the shape of the full dataset used during each run of the \
        plugin (i.e. ignore extra dimensions due to parameter tuning). """
        return self.pre_tuning_shape if self.pre_tuning_shape else\
            self.data_obj.get_shape()

    def __check_dimensions(self, indices, core_dir, slice_dir, nDims):
        if len(indices) is not len(slice_dir):
            sys.exit("Incorrect number of indices specified when accessing "
                     "data.")

        if (len(core_dir)+len(slice_dir)) is not nDims:
            sys.exit("Incorrect number of data dimensions specified.")

    def __set_slice_dimensions(self, first_sdim=None):
        """ Set the slice dimensions in the pluginData meta data dictionary.\
        Reorder pattern slice_dims to ensure first_sdim is at the front.
        """
        pattern = self.data_obj.get_data_patterns()[self.get_pattern_name()]
        slice_dims = pattern['slice_dims']

        if first_sdim:
            slice_dims = list(slice_dims)
            first_sdim = \
                self.data_obj.get_data_dimension_by_axis_label(first_sdim)
            slice_dims.insert(0, slice_dims.pop(slice_dims.index(first_sdim)))
            pattern['slice_dims'] = tuple(slice_dims)

        self.meta_data.set('slice_dims', tuple(slice_dims))

    def get_slice_dimension(self):
        """
        Return the position of the slice dimension in relation to the data
        handed to the plugin.
        """
        core_dirs = self.data_obj.get_core_dimensions()
        slice_dir = self.data_obj.get_slice_dimensions()[0]
        return list(set(core_dirs + (slice_dir,))).index(slice_dir)

    def get_data_dimension_by_axis_label(self, label, contains=False):
        """
        Return the dimension of the data in the plugin that has the specified
        axis label.
        """
        label_dim = self.data_obj.get_data_dimension_by_axis_label(
                label, contains=contains)
        plugin_dims = self.data_obj.get_core_dimensions()
        if self._get_max_frames_process() > 1 or self.max_frames == 'multiple':
            plugin_dims += (self.get_slice_dimension(),)
        return list(set(plugin_dims)).index(label_dim)

    def set_slicing_order(self, order):  # should this function be deleted?
        """
        Reorder the slice dimensions.  The fastest changing slice dimension
        will always be the first one stated in the pattern key ``slice_dir``.
        The input param is a tuple stating the desired order of slicing
        dimensions relative to the current order.
        """
        slice_dirs = self.data_obj.get_slice_dimensions()
        if len(slice_dirs) < len(order):
            raise Exception("Incorrect number of dimensions specifed.")
        ordered = [slice_dirs[o] for o in order]
        remaining = [s for s in slice_dirs if s not in ordered]
        new_slice_dirs = tuple(ordered + remaining)
        self.get_pattern()['slice_dir'] = new_slice_dirs

    def get_core_dimensions(self):
        """
        Return the position of the core dimensions in relation to the data
        handed to the plugin.
        """
        core_dims = self.data_obj.get_core_dimensions()
        first_slice_dim = (self.data_obj.get_slice_dimensions()[0],)
        plugin_dims = np.sort(core_dims + first_slice_dim)
        return np.searchsorted(plugin_dims, np.sort(core_dims))

    def set_fixed_dimensions(self, dims, values):
        """ Fix a data direction to the index in values list.

        :param list(int) dims: Directions to fix
        :param list(int) value: Index of fixed directions
        """
        slice_dirs = self.data_obj.get_slice_dimensions()
        if set(dims).difference(set(slice_dirs)):
            raise Exception("You are trying to fix a direction that is not"
                            " a slicing direction")
        self.meta_data.set("fixed_dimensions", dims)
        self.meta_data.set("fixed_dimensions_values", values)
        self.__set_slice_dimensions()
        shape = list(self.data_obj.get_shape())
        for dim in dims:
            shape[dim] = 1
        self.data_obj.set_shape(tuple(shape))
        #self.__set_shape()

    def _get_fixed_dimensions(self):
        """ Get the fixed data directions and their indices

        :returns: Fixed directions and their associated values
        :rtype: list(list(int), list(int))
        """
        fixed = []
        values = []
        if 'fixed_dimensions' in self.meta_data.get_dictionary():
            fixed = self.meta_data.get("fixed_dimensions")
            values = self.meta_data.get("fixed_dimensions_values")
        return [fixed, values]

    def _get_data_slice_list(self, plist):
        """ Convert a plugin data slice list to a slice list for the whole
        dataset, i.e. add in any missing dimensions.
        """
        nDims = len(self.get_shape())
        all_dims = self.get_core_dimensions() + self.get_slice_dimension()
        extra_dims = all_dims[nDims:]
        dlist = list(plist)
        for i in extra_dims:
            dlist.insert(i, slice(None))
        return tuple(dlist)

    def _get_max_frames_process(self):
        """ Get the number of frames to process for each run of process_frames.

        If the number of frames is not divisible by the previewing ``chunk``
        value then amend the number of frames to gcd(frames, chunk)

        :returns: Number of frames to process
        :rtype: int
        """
        if self._plugin and self._plugin.chunk > 1:
            frame_chunk = self.meta_data.get("max_frames_process")
            chunk = self.data_obj.get_preview().get_starts_stops_steps(
                key='chunks')[self.get_slice_directions()[0]]
            self.meta_data.set('max_frames_process', gcd(frame_chunk, chunk))
        return self.meta_data.get("max_frames_process")

    def _get_max_frames_transfer(self):
        """ Get the number of frames to transfer for each run of
        process_frames. """
        return self.meta_data.get('max_frames_transfer')

    def _set_no_squeeze(self):
        self.no_squeeze = True

    def _get_no_squeeze(self):
        return self.no_squeeze
    
    def _set_rank_inc(self, n):
        """ Increase the rank of the array passed to the plugin by n.
        
        :param int n: Rank increment.
        """
        self._increase_rank = n
    
    def _get_rank_inc(self):
        """ Return the increased rank value
        
        :returns: Rank increment
        :rtype: int
        """
        return self._increase_rank

    def _set_meta_data(self):
        fixed, _ = self._get_fixed_dimensions()
        sdir = \
            [s for s in self.data_obj.get_slice_dimensions() if s not in fixed]
        shape = self.data_obj.get_shape()
        shape_before_tuning = self._get_shape_before_tuning()

        diff = len(shape) - len(shape_before_tuning)
        if diff:
            shape = shape_before_tuning
            sdir = sdir[:-diff]

        if 'fix_total_frames' in list(self.meta_data.get_dictionary().keys()):
            frames = self.meta_data.get('fix_total_frames')
        else:
            frames = np.prod([shape[d] for d in sdir])

        base_names = [p.__name__ for p in self._plugin.__class__.__bases__]
        processes = self.data_obj.exp.meta_data.get('processes')

        if 'GpuPlugin' in base_names:
            n_procs = len([n for n in processes if 'GPU' in n])
        else:
            n_procs = len(processes)

        # Fixing f_per_p to be just the first slice dimension for now due to
        # slow performance from HDF5 when not slicing multiple dimensions
        # concurrently
        #f_per_p = np.ceil(frames/n_procs)
        f_per_p = np.ceil(shape[sdir[0]]/n_procs)
        self.meta_data.set('shape', shape)
        self.meta_data.set('sdir', sdir)
        self.meta_data.set('total_frames', frames)
        self.meta_data.set('mpi_procs', n_procs)
        self.meta_data.set('frames_per_process', f_per_p)
        frame_shape, b_per_f = self.set_bytes_per_frame()
        self.meta_data.set('bytes_per_frame', b_per_f)
        self.meta_data.set('bytes_per_process', b_per_f*f_per_p)
        self.meta_data.set('frame_shape', frame_shape)

    def __log_max_frames(self, mft, mfp, check=True):
        logging.debug("Setting max frames transfer for plugin %s to %d" %
                      (self._plugin, mft))
        logging.debug("Setting max frames process for plugin %s to %d" %
                      (self._plugin, mfp))
        self.meta_data.set('max_frames_process', mfp)
        if check:
            self.__check_distribution(mft)
        # (((total_frames/mft)/mpi_procs) % 1)

    def __check_distribution(self, mft):
        warn_threshold = 0.85
        nprocs = self.meta_data.get('mpi_procs')
        nframes = self.meta_data.get('total_frames')
        temp = (((nframes/mft)/float(nprocs)) % 1)
        if temp != 0.0 and temp < warn_threshold:
            shape = self.meta_data.get('shape')
            sdir = self.meta_data.get('sdir')
            logging.warning('UNEVEN FRAME DISTRIBUTION: shape %s, nframes %s ' +
                         'sdir %s, nprocs %s', shape, nframes, sdir, nprocs)

    def _set_padding_dict(self):
        if self.padding and not isinstance(self.padding, Padding):
            self.pad_dict = copy.deepcopy(self.padding)
            self.padding = Padding(self)
            for key in list(self.pad_dict.keys()):
                getattr(self.padding, key)(self.pad_dict[key])

    def plugin_data_setup(self, pattern, nFrames, split=None, slice_axis=None,
                          getall=None):
        """ Setup the PluginData object.

        :param str pattern: A pattern name
        :param int nFrames: How many frames to process at a time.  Choose from
            'single', 'multiple', 'fixed_multiple' or an integer (an integer
            should only ever be passed in exceptional circumstances)
        :keyword str slice_axis: An axis label associated with the fastest
            changing (first) slice dimension.
        :keyword list[pattern, axis_label] getall: A list of two values.  If
        the requested pattern doesn't exist then use all of "axis_label"
        dimension of "pattern" as this is equivalent to one slice of the
        original pattern.
        """

        if pattern not in self.data_obj.get_data_patterns() and getall:
            pattern, nFrames = self.__set_getall_pattern(getall, nFrames)

        # slice_axis is first slice dimension
        self.__set_pattern(pattern, first_sdim=slice_axis)
        if isinstance(nFrames, list):
            nFrames, self._frame_limit = nFrames
        self.max_frames = nFrames
        self.split = split

    def __set_getall_pattern(self, getall, nFrames):
        """ Set framework changes required to get all of a pattern of lower
        rank.
        """
        pattern, slice_axis = getall
        dim = self.data_obj.get_data_dimension_by_axis_label(slice_axis)
        # ensure data remains the same shape when 'getall' dim has length 1
        self._set_no_squeeze()
        if nFrames == 'multiple' or (isinstance(nFrames, int) and nFrames > 1):
            self._set_rank_inc(1)
        nFrames = self.data_obj.get_shape()[dim]
        return pattern, nFrames

    def plugin_data_transfer_setup(self, copy=None, calc=None):
        """ Set up the plugin data transfer frame parameters.
        If copy=pData (another PluginData instance) then copy """
        chunks = \
            self.data_obj.get_preview().get_starts_stops_steps(key='chunks')

        if not copy and not calc:
            mft, mft_shape, mfp = self._calculate_max_frames()
        elif calc:
            max_mft = calc.meta_data.get('max_frames_transfer')
            max_mfp = calc.meta_data.get('max_frames_process')
            max_nProc = int(np.ceil(max_mft/float(max_mfp)))
            nProc = max_nProc
            mfp = 1 if self.max_frames == 'single' else self.max_frames
            mft = nProc*mfp
            mft_shape = self._set_shape_transfer(self.__get_slice_size(mft))
        elif copy:
            mft = copy._get_max_frames_transfer()
            mft_shape = self._set_shape_transfer(self.__get_slice_size(mft))
            mfp = copy._get_max_frames_process()

        self.__set_max_frames(mft, mft_shape, mfp)

        if self._plugin and mft \
                and (chunks[self.data_obj.get_slice_dimensions()[0]] % mft):
            self._plugin.chunk = True
        self.__set_shape()

    def _calculate_max_frames(self):
        nFrames = self.max_frames
        self.__perform_checks(nFrames)
        td = self.data_obj._get_transport_data()
        mft, size_list = td._calc_max_frames_transfer(nFrames)
        self.meta_data.set('size_list', size_list)
        mfp = td._calc_max_frames_process(nFrames)
        if mft:
            mft_shape = self._set_shape_transfer(list(size_list))
        return mft, mft_shape, mfp

    def __set_max_frames(self, mft, mft_shape, mfp):
        self.meta_data.set('max_frames_transfer', mft)
        self.meta_data.set('transfer_shape', mft_shape)
        self.meta_data.set('max_frames_process', mfp)
        self.__log_max_frames(mft, mfp)
        # Retain the shape if the first slice dimension has length 1
        if mfp == 1 and self.max_frames == 'multiple':
            self._set_no_squeeze()

    def _get_plugin_data_size_params(self):
        nBytes = self.data_obj.get_itemsize()
        frame_shape = self.meta_data.get('frame_shape')
        total_frames = self.meta_data.get('total_frames')
        tbytes = nBytes*np.prod(frame_shape)*total_frames

        params = {'nBytes': nBytes, 'frame_shape': frame_shape,
                  'total_frames': total_frames, 'transfer_bytes': tbytes}
        return params

    def __perform_checks(self, nFrames):
        options = ['single', 'multiple']
        if not np.issubdtype(type(nFrames), np.int64) and nFrames not in options:
            e_str = ("The value of nFrames is not recognised.  Please choose "
            + "from 'single' and 'multiple' (or an integer in exceptional "
            + "circumstances).")
            raise Exception(e_str)

    def get_frame_limit(self):
        return self._frame_limit

    def get_current_frame_idx(self):
        """ Returns the index of the frames currently being processed.
        """
        global_index = self._plugin.get_global_frame_index()
        count = self._plugin.get_process_frames_counter()
        mfp = self.meta_data.get('max_frames_process')
        start = global_index[count]*mfp
        index = np.arange(start, start + mfp)
        nFrames = self.get_total_frames()
        index[index >= nFrames] = nFrames - 1
        return index



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: dxchange_loader
   :platform: Unix
   :synopsis: Load tomography data in dxchange format.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np

from savu.data.data_structures.data_types.data_plus_darks_and_flats import \
    NoImageKey
from savu.plugins.loaders.full_field_loaders.nxtomo_loader import NxtomoLoader
from savu.plugins.utils import register_plugin


@register_plugin
class DxchangeLoader(NxtomoLoader):
    """
    A class to load tomography data from a hdf5 file

    :param data_path: Path to the data. Default: 'exchange/data'.
    :param dark: dark data path and scale \
        value. Default: ['exchange/data_dark', 1].
    :param flat: flat data path and scale \
        value. Default: ['exchange/data_white', 1].
    :u*param logfile: path to the log file. Default: None.

    :*param angles: Hidden. Default: [1, 2, 3].
    :~param image_key_path: Not required Default: None.
    """

    def __init__(self, name='DxchangeLoader'):
        super(DxchangeLoader, self).__init__(name)

    def _set_dark_and_flat(self, data_obj):
        data_obj.data = NoImageKey(data_obj, None, 0)
        self._set_data(data_obj, 'flat', data_obj.data._set_flat_path)
        self._set_data(data_obj, 'dark', data_obj.data._set_dark_path)

    def _set_data(self, data_obj, name, func):
        entry, scale = self.parameters[name]
        ffile = data_obj.backing_file
        func(ffile[entry])
        data_obj.data._set_scale(name, scale)

    def _set_rotation_angles(self, data_obj):
        # set parameters here
        with open(self.parameters['logfile'], 'r') as fp:
            for i, line in enumerate(fp):
                if i == 23:
                    n_proj = int(line.split(':')[1].strip())
        angles = np.linspace(0, 180, n_proj)
        data_obj.meta_data.set("rotation_angle", angles)
        return len(angles)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: spectrum_crop
   :platform: Unix
   :synopsis: A plugin to crop a spectra

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""
import logging
import numpy as np
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class SpectrumCrop(BaseFilter, CpuPlugin):
    """
    crops a spectrum to a range

    :param crop_range: range to crop to. Default: [2., 18.].
    :param axis: range to crop to. Default: "energy".

    """

    def __init__(self):
        logging.debug("cropping spectrum")
        super(SpectrumCrop, self).__init__("SpectrumCrop")

    def process_frames(self, data):
        data = data[0]
        data_cropped = data[self.new_idx]
        return data_cropped

    def setup(self):
        in_dataset, out_datasets = self.get_datasets()
        cropped = out_datasets[0]
        cropped.create_dataset(in_dataset[0])
        in_meta = in_dataset[0].meta_data
        out_meta = cropped.meta_data
        crop_range = self.parameters['crop_range']
#         axis_labels = in_dataset[0].get_axis_label_keys()
        raw_axis = in_meta.get(self.parameters['axis'])
        logging.debug("the raw axis is: %s" % str(raw_axis))
        self.new_idx = (raw_axis>crop_range[0]) & (raw_axis<crop_range[1])
        new_axis = raw_axis[self.new_idx]
        out_meta.set(self.parameters['axis'], new_axis)
        new_len = len(new_axis)
#         print "newlen is"+str(new_len)
        shape = in_dataset[0].get_shape()[0:-1]+(new_len,)
#         print shape
        cropped.set_shape(shape)
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SPECTRUM', self.get_max_frames())
        out_pData[0].plugin_data_setup('SPECTRUM', self.get_max_frames())

    def get_max_frames(self):
        return 'single'

    def nOutput_datasets(self):
        return 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_driver
   :platform: Unix
   :synopsis: Base class or all driver plugins

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import logging
from mpi4py import MPI


class BasicDriver(object):
    """
    The base class from which all plugins should inherit.
    """

    def __init__(self):
        super(BasicDriver, self).__init__()

    def get_mem_multiply(self):
        return 1

    def _run_plugin_instances(self, transport, communicator=MPI.COMM_WORLD):
        self.__set_communicator(communicator)
        logging.info("%s.%s", self.__class__.__name__, 'pre_process')
        self.base_pre_process()
        self.pre_process()

        msg = "Pre-process completed for %s" % self.__class__.__name__
        self.plugin_barrier(msg=msg)

        logging.info("%s.%s", self.__class__.__name__, 'process_frames')
        transport._transport_process(self)

        msg = "Process_frames completed for %s" % self.__class__.__name__
        self.plugin_barrier(msg=msg)

        logging.info("%s.%s", self.__class__.__name__, 'post_process')
        self.post_process()
        self.base_post_process()

    def __set_communicator(self, comm):
        self._communicator = comm

    def get_communicator(self):
        return self._communicator

    def plugin_barrier(self, msg=''):
        return self.exp._barrier(communicator=self.get_communicator(), msg=msg)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_template
   :platform: Unix
   :synopsis: A template to create a simple plugin that takes one dataset as\
   input and returns a similar dataset as output.

.. moduleauthor:: Developer Name <email@address.ac.uk>

"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class PluginTemplate1(Plugin, CpuPlugin):
    """
    A simple plugin template with one in_dataset and one out_dataset with\
    similar characteristics, e.g. median filter.

    :param example: Example of a plugin parameter. Default: None.
    """

    def __init__(self):
        super(PluginTemplate1, self).__init__('PluginTemplate1')

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[0].plugin_data_setup('SINOGRAM', 'single')

    def pre_process(self):
        pass

    def process_frames(self, data):
        # do some processing here
        return data[0]

    def post_process(self):
        pass




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: meta_data
   :platform: Unix
   :synopsis: Contains the MetaData class which holds all information \
   required by the pipeline throughout processing.  An instance of MetaData \
   is held by the Experiment class.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import copy
import logging
from collections import OrderedDict


class MetaData(object):
    """
    The MetaData class creates a dictionary of all meta data which can be \
    accessed using the get and set methods. It also holds an instance of \
    PluginList.
    """

    def __init__(self, options={}, ordered=False):
        self.dict = OrderedDict(options) if ordered else options.copy()

    def set(self, name, value):
        """ Create and set an entry in the meta data dictionary.

        :param name: dictionary key(s). If ``name`` is a list then each
            successive name will become an entry in the dictionary which has
            the previous name as its key.
        :type name: str or list(str)
        :param value value: dictionary value

        For example,

            >>> MetaDataObj.set(['name1', 'name2'], 3)
            >>> MetaDataObj.get_dictionary()
            {'name1': {'name2': 3}}
        """
        maplist = name if isinstance(name, list) else [name]
        self.get(maplist[:-1], True)[maplist[-1]] = value

    def get(self, maplist, setFlag=False, value=True, units=False):
        """ Get a value from the meta data dictionary, given its key(s).

        :params maplist: Dictionary key(s).
        :type maplist: str or list(str)
        :returns: Value from the dictionary corresponding to the given key(s)
        :rtype: value

        Dictionaries within dictionaries are accessed by placing successive
        keys in a list.
        """
        if not maplist:
            return self.dict

        function = lambda k, d: d[k]
        maplist = (maplist if type(maplist) is list else [maplist])
        it = iter(maplist)
        accum_value = self.dict
        for x in it:
            while True:
                try:
                    accum_value = function(x, accum_value)
                except KeyError:
                    if setFlag is True:
                        accum_value[x] = {}
                        continue
                    else:
                        errorStr = 'The metadata ' + str(maplist) + \
                                   ' does not exist'
                        raise KeyError(errorStr)
                break

        if isinstance(accum_value, dict) and accum_value:
            options = OrderedDict([('value', value), ('units', units)])
            if not set(accum_value.keys()).difference(set(options.keys())):
                accum_value = [accum_value[k] for k, v in options.items()
                               if v is True]
        return accum_value

    def delete(self, entry):
        """ Delete an entry from the meta data dictionary.

        :param str entry: The dictionary key entry to delete.
        """
        try:
            del self.get_dictionary()[entry]
        except KeyError:
            logging.warning("Trying to delete a dictionary entry that doesn't "
                         "exist.")

    def get_dictionary(self):
        """ Get the meta_data dictionary.

        :returns: A dictionary.
        :rtype: dict
        """
        return self.dict

    def _set_dictionary(self, ddict):
        """ Set the meta data dictionary """
        self.dict = copy.deepcopy(ddict)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Umpa
   :platform: Unix
   :synopsis: Speckle matching using the UMPA method

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>

"""
import logging
import numpy as np

from speckle_matching import match_speckles
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

@register_plugin
class Umpa(BaseFilter, CpuPlugin):
    """
    A plugin to perform speckle tracking using the UMPA method
    :param in_datasets: A list of the dataset(s) to process. Default: ['signal','reference'].
    :param out_datasets: A list of the dataset(s) to process. Default: ['T','dx','dy','df','f'].
    :param Nw: 2*Nw + 1 is the width of the window. Default: 4.
    :param step: perform the analysis on every other _step_ pixels in both directions. Default: 1.
    :param max_shift: Do not allow shifts larger than this number of pixels. Default: 4.
    """

    def __init__(self):
        logging.debug("Starting Umpa speckle matching")
        super(Umpa,
              self).__init__("Umpa")

    def process_frames(self, data):
        signals = data[0]
        references = data[1]
        out_dict = match_speckles(signals, references, self.parameters['Nw'], self.parameters['step'], self.parameters['max_shift'], df=True, printout=False)
        return [out_dict['T'],out_dict['dx'],out_dict['dy'],out_dict['df'],out_dict['f']]

    def setup(self):
        # get all in and out datasets required by the plugin
        in_datasets, out_datasets = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        
        ## get the output shape
        Ns = self.parameters['max_shift']
        Nw = self.parameters['Nw']
        step = self.parameters['step']
        inshape = in_datasets[0].get_shape()
        imshape = inshape[-2:]
        restshape = inshape[:-3]
        ROIx = np.arange(Ns+Nw, imshape[0]-Ns-Nw-1, step)
        ROIy = np.arange(Ns+Nw, imshape[1]-Ns-Nw-1, step)


        outshape = restshape + (len(ROIx), len(ROIy))
        allDims = list(range(len(outshape)))
        in_pData[0].plugin_data_setup(self.get_plugin_pattern(), self.get_max_frames())
        in_pData[1].plugin_data_setup(self.get_plugin_pattern(), self.get_max_frames())
        pattern_list = list(in_datasets[0].get_data_patterns().keys())
        pattern_dict = in_datasets[0].get_data_patterns()
        pattern_list.remove(self.get_plugin_pattern())
        proj_in_core_dirs= np.array(in_pData[0].get_core_dimensions())
        first_core_dir = proj_in_core_dirs[0] # should be the scan index
        outlabels = in_datasets[0].get_axis_labels()
        _unused = outlabels.pop(first_core_dir)
        out_datasets[0].create_dataset(axis_labels=outlabels, shape=outshape)
        out_datasets[1].create_dataset(axis_labels=outlabels, shape=outshape)
        out_datasets[2].create_dataset(axis_labels=outlabels, shape=outshape)
        out_datasets[3].create_dataset(axis_labels=outlabels, shape=outshape)
        out_datasets[4].create_dataset(axis_labels=outlabels, shape=outshape)
        
        for pattern in pattern_list:
            core_dir = pattern_dict[pattern]['core_dims']
            core_dir = [ix-1 if ix > first_core_dir else ix for ix in core_dir]
            slice_dir = list(set(allDims)-set(core_dir))
            slice_dir = [ix-1 if ix > first_core_dir else ix for ix in slice_dir]
            dim_info = {'core_dims': core_dir, 'slice_dims': slice_dir}
            
            for dataset in out_datasets:
                dataset.add_pattern(pattern, **dim_info)



        out_pData[0].plugin_data_setup('PROJECTION', self.get_max_frames())
        out_pData[1].plugin_data_setup('PROJECTION', self.get_max_frames())
        out_pData[2].plugin_data_setup('PROJECTION', self.get_max_frames())
        out_pData[3].plugin_data_setup('PROJECTION', self.get_max_frames())
        out_pData[4].plugin_data_setup('PROJECTION', self.get_max_frames())


    def get_plugin_pattern(self):
        return "4D_SCAN"

    def nInput_datasets(self):
        return 2

    def nOutput_datasets(self):
        return 5

    def get_max_frames(self):
        return 1


# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: cpu_plugin
   :platform: Unix
   :synopsis: Base class for all plugins which use a CPU on the target machine

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.driver.plugin_driver import PluginDriver
from savu.plugins.driver.basic_driver import BasicDriver

import os
_base = BasicDriver if os.environ['savu_mode'] == 'basic' else PluginDriver


class CpuPlugin(_base):
    """
    The base class from which all plugins should inherit.
    """

    def __init__(self):
        super(CpuPlugin, self).__init__()

    def _run_plugin(self, exp, transport):
        self._run_plugin_instances(transport)
        return



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: ica
   :platform: Unix
   :synopsis: A plugin to fit peaks

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.utils import register_plugin
from savu.plugins.ptychography.base_ptycho import BasePtycho
import numpy as np


@register_plugin
class DummyPtycho(BasePtycho):
    """
    This plugin performs ptychography using the ptypy package
    """

    def __init__(self):
        super(DummyPtycho, self).__init__("DummyPtycho")

    def process_frames(self, data):
        data = data[0]
        probe = data[0]
        #print "probe is "+str(probe.shape)
        object_transmission = np.random.random(self.obj_shape).squeeze()
        positions = self.get_positions()
        return [probe, object_transmission, positions]#] add fourier error, realspace error



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. currentmodule:: savu

The Base level for Savu

use with :

import savu

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""


import os
import sys
from . import test

savuPath = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(savuPath + "/../lib"))
os.environ['savu_mode'] = 'hdf5'


def run_full_tests():
    import unittest

    print("Tests will run shortly, and may take some time to complete")
    print("The tests may raise errors, please don't worry about these as "
          "they may be raised deliberately.")
    print("The key information is in the final test results")

    path = os.path.split(test.travis.__file__)[0]
    suite = unittest.defaultTestLoader.discover(path, pattern='*test.py')
    unittest.TextTestRunner(buffer=True).run(suite)


def run_tests():
    import unittest
    from savu.test.travis.plugin_tests.reconstruction_tests.tomo_pipeline_preview_test \
        import TomoPipelinePreviewTest
    print("Running a quick test...")

    suite = \
        unittest.TestLoader().loadTestsFromTestCase(TomoPipelinePreviewTest)
    unittest.TextTestRunner(verbosity=1, buffer=True).run(suite)
    print("Test complete...")



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_multi_modal_loader
   :platform: Unix
   :synopsis: Contains a base class from which all multi-modal loaders are \
   derived.
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""

import h5py
import logging

from savu.plugins.loaders.base_loader import BaseLoader


class BaseMultiModalLoader(BaseLoader):
    """
    This class provides a base for all multi-modal loaders
    """

    def __init__(self, name='BaseMultiModalLoader'):
        super(BaseMultiModalLoader, self).__init__(name)
        self._mtype = None

    def multi_modal_setup(self, ltype, data_str, name, patterns=True):
        data_obj = self._get_file_handle(name, ltype)
        NXdef = 'NXstxm' if ltype == 'NXmonitor' else ltype
        entry = self.get_NXapp(NXdef, data_obj.backing_file, 'entry1/')[0]

        path = ('/').join([entry.name, data_str])
        data_obj.data = data_obj.backing_file[path]
        data_obj.set_shape(data_obj.data.shape)
        self._check_for_monitor_data(data_obj, entry)
        self.set_motors(data_obj, entry, 'fluo')
        if patterns:
            self.add_patterns_based_on_acquisition(data_obj, 'fluo')
        return data_obj, entry

    def _get_file_handle(self, name, ltype):
        exp = self.exp
        data_obj = exp.create_data_object("in_data", name)
        data_obj.backing_file = h5py.File(exp.meta_data.get("data_file"), 'r')
        logging.debug("Creating file '%s' '%s'_entry",
                      data_obj.backing_file.filename, ltype)
        return data_obj

    def _check_for_monitor_data(self, data_obj, entry):
        if entry.name + '/monitor/data' in data_obj.backing_file:
            control = data_obj.backing_file[entry.name + '/monitor/data']
            self.exp.meta_data.set('control', control)
            logging.debug('adding the ion chamber to the meta data')
        else:
            logging.warning('No ion chamber information. Leaving this blank')

    def set_motors(self, data_obj, entry, ltype):
        labels = [e.decode("ascii") for e in entry['data'].attrs['axes']]
        motors = [entry['data/' + e] for e in labels]
        units = self._get_attrs(motors, 'units', 'unit')
        self._mtype = self._get_attrs(motors, 'transformation_type', 'None')
        self._set_axis_labels(data_obj, motors, labels, units)

    def _get_attrs(self, entries, key, default):
        return [e.attrs[key].decode("ascii") if key in list(e.attrs.keys()) else
                default for e in entries]

    def _set_axis_labels(self, dObj, motors, labels, units):
        trans = self.get_motor_dims('translation')
        rot_dim = self._mtype.index('rotation')
        angles = motors[rot_dim][()]
        labels[rot_dim] = 'rotation_angle'
        self._set_meta_data(dObj, 'rotation_angle', angles)
        if trans:
            self._set_meta_data(dObj, 'x', motors[trans[-1]][()])
            labels[trans[-1]] = 'x'
            if len(trans) > 1:
                self._set_meta_data(dObj, 'y', motors[trans[-2]][()])
                labels[trans[-2]] = 'y'
        labels = [labels[i] + '.' + units[i] for i in range(len(labels))]
        dObj.set_axis_labels(*tuple(labels))

    def _set_meta_data(self, dObj, key, value):
        value = value[:, 0] if value.ndim > 1 else value
        dObj.meta_data.set(key, value)

    def get_motor_dims(self, key):
        return [i for i in range(len(self._mtype)) if self._mtype[i] == key]

    def add_patterns_based_on_acquisition(self, data_obj, ltype):
        dims = list(range(len(self._mtype)))

        proj_dims = tuple(self.get_motor_dims('translation'))
        if proj_dims:
            data_obj.add_pattern("PROJECTION", core_dims=proj_dims,
                                 slice_dims=tuple(set(dims) - set(proj_dims)))

        rot_dim = tuple(self.get_motor_dims('rotation'))
        if rot_dim:
            sino_dir = rot_dim + (proj_dims[-1],) if proj_dims else rot_dim
            sino_slice_dir = tuple(set(dims) - set(sino_dir))
            data_obj.add_pattern("SINOGRAM", core_dims=sino_dir,
                                 slice_dims=sino_slice_dir)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: data_with_darks_and_flats
   :platform: Unix
   :synopsis: A class for loading data that has associated dark and flat \
       fields.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np
import copy

from savu.data.data_structures.data_types.base_type import BaseType


class DataWithDarksAndFlats(BaseType):

    def __init__(self, data_obj, proj_dim, image_key):
        super(DataWithDarksAndFlats, self).__init__()
        self.fscale = 1
        self.dscale = 1
        self.flat_updated = False
        self.dark_updated = False
        self.data = data_obj.data
        self.dark_flat_slice_list = []
        self.dtype = data_obj.data.dtype

    def _base_extra_params(self):
        """ global class parameter names that are updated outside of __init__
        """
        extras = ['fscale', 'dscale', 'dark_updated', 'flat_updated',
                  'dark_flat_slice_list']
        return extras

    def _override_data_type(self, data):
        self.data = data

    def get_image_key(self):
        preview_sl = self.data_obj.get_preview()._get_preview_slice_list()
        if preview_sl is None:
            return self.image_key
        return self.__get_preview_image_key(preview_sl[self.proj_dim])

    def _get_image_key_data_shape(self):
        data_idx = self.get_index(0)
        new_shape = list(self.data.shape)
        new_shape[self.proj_dim] = len(data_idx)
        return tuple(new_shape)

    def _getitem_imagekey(self, idx):
        index = list(idx)
        index[self.proj_dim] = \
            self.get_index(0, full=True)[idx[self.proj_dim]].tolist()
        return self.data[tuple(index)]

    def _getitem_noimagekey(self, idx):
        return self.data[idx]

    def __setitem__(self, key, val):
        self.data[key] = val

    def get_dark_flat_slice_list(self):
        slice_list = self.data_obj._preview._get_preview_slice_list()
        remove_dim = \
            self.data_obj.get_data_dimension_by_axis_label('rotation_angle')
        slice_list[remove_dim] = slice(None)

        if len(slice_list) > 3:
            idx = np.arange(0, len(slice_list))
            detX = self.data_obj.get_data_dimension_by_axis_label('detector_x')
            detY = self.data_obj.get_data_dimension_by_axis_label('detector_y')
            remove = set(idx).difference({remove_dim, detX, detY})
            for dim in sorted(list(remove), reverse=True):
                del slice_list[dim]

        return slice_list

    def _set_scale(self, name, scale):
        self.set_flat_scale(scale) if name == 'flat' else \
            self.set_dark_scale(scale)

    def set_flat_scale(self, fscale):
        self.fscale = float(fscale)

    def set_dark_scale(self, dscale):
        self.dscale = float(dscale)

    def get_shape(self):
        return self.shape

    def dark_mean(self):
        """ Get the averaged dark projection data. """
        return self._calc_mean(self.dark())

    def flat_mean(self):
        """ Get the averaged flat projection data. """
        return self._calc_mean(self.flat())

    def _calc_mean(self, data):
        return data if len(data.shape) is 2 else \
            data.mean(self.proj_dim).astype(np.float32)

    def get_index(self, key, full=False):
        """ Get the projection index of a specific image key value.

        :params int key: the image key value
        """
        if full is True:
            return np.where(self.image_key == key)[0]
        else:
            return self.__get_preview_index(key)

    def __get_preview_index(self, key):
        # Remove this try/except statement
        preview_sl = self.data_obj.get_preview()._get_preview_slice_list()
        if preview_sl is None:
            return np.where(self.image_key == key)[0]
        else:
            return self.__get_reduced_index(key, preview_sl[self.proj_dim])

    def __get_preview_image_key(self, slice_list):
        # all data entries
        data_idx = np.where(self.image_key == 0)[0]
        preview_idx = np.arange(len(data_idx))[slice_list]
        # check the inconsistency regarding the preview of angles, e.g [0:10,:,:]
        if len(data_idx) == len(preview_idx):
            remove_idx = np.delete(data_idx, preview_idx[::-1])
        else:
            remove_idx = []
        return np.delete(self.image_key, remove_idx)

    def __get_reduced_index(self, key, slice_list):
        """ Get the projection index of a specific image key value when\
            previewing has been applied. """
        preview_image_key = self.__get_preview_image_key(slice_list)
        return np.where(preview_image_key == key)[0]

    def __get_data(self, key):
        index = [slice(None)] * self.nDims
        rot_dim = self.data_obj.get_data_dimension_by_axis_label(
            'rotation_angle')

        # separate the transfer of data for slice lists with entries far \
        # apart, as this significantly improves hdf5 performance.
        split_diff = 10
        k_idx = self.get_index(key)
        if not k_idx.size:
            return np.array([])

        k_idx = np.split(k_idx, np.where(np.diff(k_idx) > split_diff)[0] + 1)

        index[self.proj_dim] = k_idx[0]
        data = self.data[tuple(index)]

        for idx in k_idx[1:]:
            index[self.proj_dim] = idx
            data = np.append(data, self.data[tuple(index)], axis=rot_dim)

        if not self.dark_flat_slice_list[key]:
            return data

        sl = list(copy.deepcopy(self.dark_flat_slice_list[key]))
        if len(data.shape) is 2:
            del sl[rot_dim]
        return data[tuple(sl)]

    def dark_image_key_data(self):
        """ Get the dark data. """
        return self.__get_data(2) * self.dscale

    def flat_image_key_data(self):
        """ Get the flat data. """
        return self.__get_data(1) * self.fscale

    def update_dark(self, data):
        self.dark_updated = data
        self.dscale = 1
        self.dark_flat_slice_list[2] = None
        self.data_obj.meta_data.set('dark', self._calc_mean(data))

    def update_flat(self, data):
        self.flat_updated = data
        self.fscale = 1
        self.dark_flat_slice_list[1] = None
        self.data_obj.meta_data.set('flat', self._calc_mean(data))

    def _set_dark_and_flat(self):
        slice_list = self.data_obj._preview._get_preview_slice_list()
        if slice_list:
            self.dark_flat_slice_list = \
                [tuple(self.get_dark_flat_slice_list())] * 3


class ImageKey(DataWithDarksAndFlats):
    """ This class is used to get data from a dataset with an image key. """

    def __init__(self, data_obj, image_key, proj_dim, ignore=None):
        self.data_obj = data_obj
        self.image_key = image_key
        self.proj_dim = proj_dim
        self.ignore = ignore
        super(ImageKey, self).__init__(data_obj, proj_dim, image_key)
        self.shape = self._get_image_key_data_shape()
        self.nDims = len(self.shape)
        self._getitem = self._getitem_imagekey
        if ignore:
            self.__ignore_image_key_entries(ignore)

    def clone_data_args(self, args, kwargs, extras):
        """ List the arguments required to clone this datatype
        """
        args = ['self', 'image_key', 'proj_dim']
        kwargs['ignore'] = 'ignore'
        extras = self._base_extra_params()
        return args, kwargs, extras

    def map_input_args(self, args, kwargs, cls, extras):
        """ List all information required to keep this data set after a plugin
        has completed (may require conversion to another type)
        """
        args = ['self', None, 'proj_dim']
        kwargs['dark'] = 'dark'
        kwargs['flat'] = 'flat'
        extras = self._base_extra_params()
        cls = NoImageKey.__module__ + '.NoImageKey'
        return args, kwargs, cls, extras

    def _finish_cloning(self):
        self.dark_flat_slice_list[0] = None

    def __getitem__(self, idx):
        return self._getitem(idx)

    def __ignore_image_key_entries(self, ignore):
        a, a, start, end = self._get_start_end_idx(self.get_index(1))
        if not isinstance(ignore, list):
            ignore = [ignore]
        for batch in ignore:
            self.image_key[start[batch - 1]:end[batch - 1] + 1] = 3

    def dark(self):
        """ Get the dark data. """
        return self.dark_updated if self.dark_updated is not False else \
            self.dark_image_key_data()

    def flat(self):
        """ Get the flat data. """
        return self.flat_updated if self.flat_updated is not False else \
            self.flat_image_key_data()


class NoImageKey(DataWithDarksAndFlats):
    """ This class is used to get data from a dataset with separate darks and
        flats. """

    def __init__(self, data_obj, image_key, proj_dim, dark=None, flat=None):
        self.data_obj = data_obj
        self.image_key = image_key
        self.proj_dim = proj_dim
        super(NoImageKey, self).__init__(data_obj, proj_dim, image_key)
        self.dark_path = dark
        self.flat_path = flat
        self.orig_image_key = copy.copy(image_key)
        self.flat_image_key = []
        self.dark_image_key = []

        # darks and flats belong to another dataset with an image key
        if self.image_key is not None:
            self.shape = self._get_image_key_data_shape()
            self._getitem = self._getitem_imagekey
        else:
            self.shape = data_obj.data.shape
            self._getitem = self._getitem_noimagekey

        self.data_obj = data_obj
        self.nDims = len(self.shape)

    def clone_data_args(self, args, kwargs, extras):
        # these are the arguments required when creating a class instance
        args = ['self', 'image_key', 'proj_dim']  # always 'self goes first'
        kwargs['dark'] = 'dark_path'
        kwargs['flat'] = 'flat_path'
        # global class parameter names that are updated outside of __init__
        extras = ['image_key', 'flat_image_key', 'flat_path',
                  'dark_image_key', 'dark_path'] + self._base_extra_params()
        return args, kwargs, extras

    def __getitem__(self, idx):
        return self._getitem(idx)

    def _post_clone_updates(self):
        self.dark_flat_slice_list[0] = None

    def _set_fake_key(self, fakekey):
        # useful if the darks and flats did belong to data with an
        # image key in a previous plugin
        self.image_key = fakekey

    def _set_flat_path(self, path, imagekey=[]):
        self.flat_image_key = imagekey
        self.flat_path = path

    def _set_dark_path(self, path, imagekey=[]):
        self.dark_image_key = imagekey
        self.dark_path = path

    def get_shape(self):
        return self.shape

    def dark(self):
        """ Get the dark data. """
        if self.dark_updated is not False:
            return self.dark_updated
        if len(self.dark_image_key) > 0:
            self.image_key = self.dark_image_key
            dark = self.dark_image_key_data()
            self.image_key = self.orig_image_key
            return dark
        return self.dark_path[self.dark_flat_slice_list[2]] * self.dscale

    def flat(self):
        """ Get the flat data. """
        if self.flat_updated is not False:
            return self.flat_updated
        if len(self.flat_image_key) > 0:
            self.image_key = self.flat_image_key
            flat = self.flat_image_key_data()
            self.image_key = self.orig_image_key
            return flat
        return self.flat_path[self.dark_flat_slice_list[1]] * self.fscale
